# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os

from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import BaseIdentityModule
from dlk.utils.import_module import import_module_dir
from dlk.utils.register import register


@cregister("embedding", "identity")
class IdentityModuleConfig:
    """identity embedding"""

    pass


@register("embedding", "identity")
class IdentityModule(BaseIdentityModule):
    pass


class EmbeddingInput(object):
    """docstring for EmbeddingInput"""

    def __init__(self, **args):
        super(EmbeddingInput, self).__init__()
        self.input_ids = args.get("input_ids", None)


class EmbeddingOutput(object):
    """docstring for EmbeddingOutput"""

    def __init__(self, **args):
        super(EmbeddingOutput, self).__init__()
        self.represent = args.get("represent", None)


embedding_dir = os.path.dirname(__file__)
import_module_dir(embedding_dir, "dlk.nn.layer.embedding")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Callable, Dict, List, Set

import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "seq_lab_relabel")
class SeqLabRelabelConfig(BaseSubProcessorConfig):
    """the sequence labeling relabel subprocessor"""

    train_data_set = ListField(
        value=["train", "valid", "test"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )

    class InputMap:
        word_ids = StrField(value="word_ids", help="the word ids")
        offsets = StrField(value="offsets", help="the offsets")
        entities_info = StrField(value="entities_info", help="the entities info")

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OutputMap:
        labels = StrField(value="labels", help="the label names")

    output_map = NestField(
        value=OutputMap,
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )
    drop = StrField(
        value="shorter",
        options=["longer", "shorter", "none"],
        help="the drop strategy for the overlap entities",
    )
    start_label = StrField(value="S", help="the start label")
    end_label = StrField(value="E", help="the end label")
    clean_droped_entity = BoolField(
        value=True, help="whether clean the dropped entity for calc metrics"
    )
    entity_priority = ListField(
        value=[],
        suggestions=[["Product", "Brand"]],
        help="the entity priority, when conflict, will keep the entity with the highest priority",
    )
    priority_trigger = IntField(
        value=1,
        help="if the overlap entity abs(length_a - length_b)<=priority_trigger, will trigger the entity_priority strategy",
    )


@register("subprocessor", "seq_lab_relabel")
class SeqLabRelabel(BaseSubProcessor):
    """
    Relabel the json data to bio
    """

    def __init__(self, stage: str, config: SeqLabRelabelConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.stage = stage
        self.config = config
        self.entity_priority = {
            entity: priority
            for priority, entity in enumerate(self.config.entity_priority)
        }

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """firstpiece relabel the data

        Args:
            data: one is like

            >>> {
            >>>     "uuid": '**-**-**-**'
            >>>     "sentence": "Mie Merah - Buah Bit",
            >>>     "offsets": see the offsets in fast_tokenizer
            >>>     "entities_info": [
            >>>                 {
            >>>                     "end": 9,
            >>>                     "start": 0,
            >>>                     "labels": [
            >>>                         "Product"
            >>>                     ]
            >>>                 },
            >>>             ]
            >>>         },
            >>>     ],
            >>> },

            deliver_meta:
                ignore
        Returns:
            relabeld data
        """
        data[
            [self.config.output_map.labels, self.config.input_map.entities_info]
        ] = data.apply(self.relabel, axis=1, result_type="expand")
        return data

    def find_position_in_offsets(
        self,
        position: int,
        offset_list: List,
        sub_word_ids: List,
        start: int,
        end: int,
        is_start: bool = False,
    ):
        """find the sub_word index which the offset_list[index][0]<=position<offset_list[index][1]

        Args:
            position: position
            offset_list: list of all tokens offsets
            sub_word_ids: word_ids from tokenizer
            start: start search index
            end: end search index
            is_start: is the position is the start of target token, if the is_start==True and cannot find return -1

        Returns:
            the index of the offset which include position

        """
        while start < end:
            if sub_word_ids[start] is None:
                start += 1
            elif position >= offset_list[start][0] and position < offset_list[start][1]:
                return start
            elif position < offset_list[start][0]:
                if start == 1 and offset_list[0] == [0, 0]:
                    return 1
                if is_start:
                    return -1
                else:
                    return start - 1
            else:
                start += 1
        return -1

    def relabel(self, one_ins: pd.Series):
        """make token label, if use the first piece label please use the 'seq_lab_firstpiece_relabel'

        Args:
            one_ins: include sentence, entity_info, offsets

        Returns:
            labels(labels for each subtoken)

        """
        pre_clean_entities_info = one_ins[self.config.input_map.entities_info]
        pre_clean_entities_info.sort(key=lambda x: x["start"])
        offsets: List = one_ins[self.config.input_map.offsets]
        sub_word_ids: List = one_ins[self.config.input_map.word_ids]
        if not sub_word_ids:
            logger.warning(
                f"entity_info: {pre_clean_entities_info}, offsets: {offsets} "
            )

        entities_info = []
        pre_end = -1
        pre_length = 0
        pre_label = ""
        for entity_info in pre_clean_entities_info:
            assert (
                len(entity_info["labels"]) == 1
            ), f"currently we just support one label for one entity"
            if entity_info["start"] < pre_end:  # if overlap will remove one
                if self.config.drop == "none":
                    pass
                elif (
                    abs(entity_info["end"] - entity_info["start"] - pre_length)
                    <= self.config.priority_trigger
                ):
                    pre_label_order = self.entity_priority.get(pre_label, 1e9)
                    label_order = self.entity_priority.get(
                        entity_info["labels"][0], 1e9
                    )
                    if label_order < pre_label_order:
                        entities_info.pop()
                    else:
                        continue
                elif self.config.drop == "shorter":
                    if entity_info["end"] - entity_info["start"] > pre_length:
                        entities_info.pop()
                    else:
                        continue
                elif self.config.drop == "longer":
                    if entity_info["end"] - entity_info["start"] < pre_length:
                        entities_info.pop()
                    else:
                        continue
                else:
                    raise PermissionError(
                        f"The drop method must in 'none'/'shorter'/'longer'"
                    )
                pre_label = entity_info["labels"][0]
            entities_info.append(entity_info)
            pre_end = entity_info["end"]
            pre_length = entity_info["end"] - entity_info["start"]

        cur_token_index = 0
        offset_length = len(offsets)
        sub_labels = []
        for entity_info in entities_info:
            start_token_index = self.find_position_in_offsets(
                entity_info["start"],
                offsets,
                sub_word_ids,
                cur_token_index,
                offset_length,
                is_start=True,
            )
            if start_token_index == -1:
                logger.warning(
                    f"cannot find the entity_info : {entity_info}, offsets: {offsets} "
                )
                continue
            for _ in range(start_token_index - cur_token_index):
                sub_labels.append("O")
            end_token_index = self.find_position_in_offsets(
                entity_info["end"] - 1,
                offsets,
                sub_word_ids,
                start_token_index,
                offset_length,
            )
            assert (
                end_token_index != -1
            ), f"entity_info: {entity_info}, offsets: {offsets}"
            sub_labels.append("B-" + entity_info["labels"][0])
            for _ in range(end_token_index - start_token_index):
                sub_labels.append("I-" + entity_info["labels"][0])
            cur_token_index = end_token_index + 1
        assert cur_token_index <= offset_length
        for _ in range(offset_length - cur_token_index):
            sub_labels.append("O")

        if sub_word_ids[0] is None:
            sub_labels[0] = self.config.start_label

        if sub_word_ids[offset_length - 1] is None:
            sub_labels[offset_length - 1] = self.config.end_label

        if len(sub_labels) != offset_length:
            logger.error(f"{len(sub_labels)} vs {offset_length}")
            for i in one_ins:
                logger.error(f"{i}")
            raise PermissionError

        if not self.config.clean_droped_entity:
            entities_info = one_ins[self.config.input_map.entities_info]
        return sub_labels, entities_info



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging

from tokenizers import Tokenizer
from tokenizers.normalizers import NFC, NFD, Lowercase, Strip, StripAccents
from tokenizers.pre_tokenizers import (
    BertPreTokenizer,
    ByteLevel,
    Whitespace,
    WhitespaceSplit,
)
from tokenizers.processors import TemplateProcessing

logger = logging.getLogger(__name__)


class TokenizerPostprocessorFactory(object):
    """docstring for TokenizerPostprocessorFactory"""

    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

    @property
    def bert(self):
        """bert postprocess

        Returns:
            bert postprocess

        """
        vocab = self.tokenizer.get_vocab()
        try:
            cls_id = vocab["[CLS]"]
            sep_id = vocab["[SEP]"]
        except Exception as e:
            # logger.error(f"`[CLS]` or `[SEP]` is not a token in this tokenizer.", )
            logger.exception(f"`[CLS]` or `[SEP]` is not a token in this tokenizer.")
            raise e
            # raise PermissionError(f"`[CLS]` or `[SEP]` is not a token in this tokenizer.")

        def _wrap():
            return TemplateProcessing(
                single="[CLS] $A [SEP]",
                pair="[CLS] $A [SEP] $B:1 [SEP]:1",
                special_tokens=[
                    ("[CLS]", cls_id),
                    ("[SEP]", sep_id),
                ],
            )

        return _wrap

    def get(self, name):
        """get postprocess by name

        Returns:
            postprocess

        """
        return self.__getattribute__(name)


class PreTokenizerFactory(object):
    """PreTokenizerFactory"""

    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

    @property
    def bytelevel(self):
        """byte level pre_tokenizer

        Returns:
            ByteLevel

        """
        return ByteLevel

    @property
    def bert(self):
        """bert pre_tokenizer

        Returns:
            BertPreTokenizer

        """
        return BertPreTokenizer

    @property
    def whitespace(self):
        """whitespace pre_tokenizer

        Returns:
            Whitespace

        """
        return Whitespace

    @property
    def whitespacesplit(self):
        """whitespacesplit pre_tokenizer

        Returns:
            WhitespaceSplit

        """
        return WhitespaceSplit

    def get(self, name):
        """get pretokenizer by name

        Returns:
            postprocess

        """
        return self.__getattribute__(name)


class TokenizerNormalizerFactory(object):
    """TokenizerNormalizerFactory"""

    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

    @property
    def lowercase(self):
        """do lowercase normalizers

        Returns:
            Lowercase

        """
        return Lowercase

    @property
    def nfd(self):
        """do nfd normalizers

        Returns:
            NFD

        """
        return NFD

    @property
    def nfc(self):
        """do nfc normalizers

        Returns:
            NFC

        """
        return NFC

    @property
    def strip_accents(self):
        """do strip normalizers

        Returns:
            StripAccents

        """
        return StripAccents

    @property
    def strip(self):
        """do strip normalizers

        Returns:
            StripAccents

        """
        return Strip

    def get(self, name):
        """get normalizers by name

        Returns:
            Normalizer

        """
        return self.__getattribute__(name)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Dict, List

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import Module


@cregister("module", "crf")
class CRFConfig:
    """the ConditionalRandomField config"""

    output_size = IntField(value=MISSING, minimum=0, help="the output size")
    batch_first = BoolField(value=True, help="whether the input is batch first")
    reduction = StrField(
        value="mean",
        options=["none", "sum", "mean", "token_mean"],
        help="the reduction method",
    )


@register("module", "crf")
class ConditionalRandomField(Module):
    """CRF, training_step for training, forward for decode。"""

    def __init__(self, config: CRFConfig):
        super(ConditionalRandomField, self).__init__()
        self.config = config

        self.num_tags = self.config.output_size

        self.transitions = nn.parameter.Parameter(
            torch.randn(self.num_tags, self.num_tags)
        )
        self.start_transitions = nn.parameter.Parameter(torch.randn(self.num_tags))
        self.end_transitions = nn.parameter.Parameter(torch.randn(self.num_tags))

    def init_weight(self, method: Callable):
        """init the weight of transitions, start_transitions and end_transitions

        Initialize the transition parameters.
        The parameters will be initialized randomly from a uniform distribution
        between -0.1 and 0.1.

        Args:
            method: init method, no use

        Returns:
            None

        """

        nn.init.normal_(self.transitions, -1, 0.1)
        nn.init.uniform_(self.start_transitions, -0.1, 0.1)
        nn.init.uniform_(self.end_transitions, -0.1, 0.1)

    def _normalizer_likelihood(self, logits: torch.FloatTensor, mask: torch.ByteTensor):
        """Computes the (batch_size,) denominator term for the log-likelihood.

        The sum of the likelihoods across all possible state sequences.

        Args:
            logits: max_len*batch_size*num_tags
            mask: max_len*batch_size

        Returns:
            batch_size*every sum

        """
        seq_len, batch_size, n_tags = logits.size()
        alpha = logits[0]
        alpha = alpha + self.start_transitions.view(1, -1)

        flip_mask = mask.eq(False)

        for i in range(1, seq_len):
            emit_score = logits[i].view(batch_size, 1, n_tags)
            trans_score = self.transitions.view(1, n_tags, n_tags)
            tmp = alpha.view(batch_size, n_tags, 1) + emit_score + trans_score
            alpha = torch.logsumexp(tmp, 1).masked_fill(
                flip_mask[i].view(batch_size, 1), 0
            ) + alpha.masked_fill(mask[i].eq(True).view(batch_size, 1), 0)

        alpha = alpha + self.end_transitions.view(1, -1)

        return torch.logsumexp(alpha, 1)

    def _gold_score(
        self, logits: torch.FloatTensor, tags: torch.LongTensor, mask: torch.ByteTensor
    ):
        """Compute the score for the gold path.

        Args:
            logits: max_len*batch_size*num_tags
            tags: max_len*batch_size
            mask: max_len*batch_size

        Returns:
            batch_size*every_gold_score

        """
        seq_len, batch_size, _ = logits.size()
        batch_idx = torch.arange(batch_size, dtype=torch.long, device=logits.device)
        seq_idx = torch.arange(seq_len, dtype=torch.long, device=logits.device)

        # trans_socre [L-1, B]
        mask = mask.eq(True)
        flip_mask = mask.eq(False)
        trans_score = self.transitions[tags[: seq_len - 1], tags[1:]].masked_fill(
            flip_mask[1:, :], 0
        )
        # emit_score [L, B]
        emit_score = logits[
            seq_idx.view(-1, 1), batch_idx.view(1, -1), tags
        ].masked_fill(flip_mask, 0)
        # score [L-1, B]
        score = trans_score + emit_score[: seq_len - 1, :]
        score = score.sum(0) + emit_score[-1].masked_fill(flip_mask[-1], 0)
        st_scores = self.start_transitions.view(1, -1).repeat(batch_size, 1)[
            batch_idx, tags[0]
        ]
        last_idx = mask.long().sum(0) - 1
        ed_scores = self.end_transitions.view(1, -1).repeat(batch_size, 1)[
            batch_idx, tags[last_idx, batch_idx]
        ]
        score = score + st_scores + ed_scores
        # return [B,]
        return score

    def training_step(
        self, logits: torch.FloatTensor, tags: torch.LongTensor, mask: torch.LongTensor
    ):
        """training step, calc the loss

        Args:
            logits: emissions, batch_size*max_len*num_tags
            tags: batch_size*max_len
            mask: batch_size*max_len, mask==0 means padding

        Returns:
            loss

        """
        logits = logits.transpose(0, 1)
        tags = tags.transpose(0, 1).long()
        mask = mask.transpose(0, 1).byte()
        all_path_score = self._normalizer_likelihood(logits, mask)
        gold_path_score = self._gold_score(logits, tags, mask)
        loss = all_path_score - gold_path_score
        return loss.mean()

    def forward(self, logits: torch.FloatTensor, mask: torch.LongTensor):
        """predict step, get the best path

        Args:
            logits: emissions, batch_size*max_len*num_tags
            mask: batch_size*max_len, mask==0 means padding

        Returns:
            batch*max_len

        """
        logits = logits.transpose(0, 1)  # L, B, H
        mask = mask.transpose(0, 1)
        return self._viterbi_decode(logits, mask)

    def _viterbi_decode(
        self, emissions: torch.FloatTensor, mask: torch.LongTensor
    ) -> torch.Tensor:
        """predict step, get the best path

        Args:
            logits: emissions, max_len*batch_size*num_tags
            mask: max_len*batch_size, mask==0 means padding

        Returns:
            batch*max_len

        """
        # emissions: (seq_length, batch_size, num_tags)
        # mask: (seq_length, batch_size)
        assert emissions.dim() == 3 and mask.dim() == 2
        assert emissions.shape[:2] == mask.shape
        assert emissions.size(2) == self.num_tags
        assert mask[0].all()

        seq_length, batch_size = mask.shape

        # Start transition and first emission
        # shape: (batch_size, num_tags)
        score = self.start_transitions + emissions[0]
        history = torch.jit.annotate(List[int], [])

        # score is a tensor of size (batch_size, num_tags) where for every batch,
        # value at column j stores the score of the best tag sequence so far that ends
        # with tag j
        # history saves where the best tags candidate transitioned from; this is used
        # when we trace back the best tag sequence

        # Viterbi algorithm recursive case: we compute the score of the best tag sequence
        # for every possible next tag
        for i in range(1, seq_length):
            # Broadcast viterbi score for every possible next tag
            # shape: (batch_size, num_tags, 1)
            broadcast_score = score.unsqueeze(2)

            # Broadcast emission score for every possible current tag
            # shape: (batch_size, 1, num_tags)
            broadcast_emission = emissions[i].unsqueeze(1)

            # Compute the score tensor of size (batch_size, num_tags, num_tags) where
            # for each sample, entry at row i and column j stores the score of the best
            # tag sequence so far that ends with transitioning from tag i to tag j and emitting
            # shape: (batch_size, num_tags, num_tags)
            next_score = broadcast_score + self.transitions + broadcast_emission

            # Find the maximum score over all possible current tag
            # shape: (batch_size, num_tags)
            next_score, indices = next_score.max(dim=1)

            # Set score to the next score if this timestep is valid (mask == 1)
            # and save the index that produces the next score
            # shape: (batch_size, num_tags)
            score = torch.where(mask.bool()[i].unsqueeze(1), next_score, score)
            history.append(indices)

        # End transition score
        # shape: (batch_size, num_tags)
        score += self.end_transitions

        # Now, compute the best path for each sample

        # shape: (batch_size,)
        seq_ends = mask.long().sum(dim=0) - 1
        best_tags_list = []
        best_tags_list = torch.jit.annotate(List[List[int]], [])

        for idx in range(batch_size):
            # Find the tag which maximizes the score at the last timestep; this is our best tag
            # for the last timestep
            _, best_last_tag = score[idx].max(dim=0)
            best_tags = [best_last_tag.item()]

            # We trace back where the best last tag comes from, append that to our best tag
            # sequence, and trace it back again, and so on
            for hist in reversed(history[: seq_ends[idx]]):
                best_last_tag = hist[idx][best_tags[-1]]
                best_tags.append(best_last_tag.item())

            # Reverse the order because we start from the last timestep
            best_tags.reverse()
            best_tags_list.append(best_tags)

        output = torch.jit.annotate(List[List[int]], [])
        for tag_list in best_tags_list:
            if len(tag_list) < seq_length:
                tag_list = tag_list + [-1] * (seq_length - len(tag_list))
            output.append(tag_list)
        return torch.tensor(output, dtype=torch.long, device=mask.device)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from logging.handlers import RotatingFileHandler


class CustomFormatter(logging.Formatter):
    cyan = "\x1b[36m"
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format_rep = "%(asctime)s - dlk -%(levelname)7s -%(message)s"

    FORMATS = {
        logging.DEBUG: grey + format_rep + reset,
        logging.INFO: cyan + format_rep + reset,
        logging.WARNING: yellow + format_rep + reset,
        logging.ERROR: red + format_rep + reset,
        logging.CRITICAL: bold_red + format_rep + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


global_logger = logging.getLogger()
log_level_map = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}
log_level = log_level_map.get(os.environ.get("DLK_LOG_LEVEL", "INFO"), "INFO")
global_logger.setLevel(log_level)


if not global_logger.hasHandlers():
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    ch.setFormatter(CustomFormatter())

    global_logger.addHandler(ch)
    global_logger.propagate = False


def logfile(
    file_name: str = "logs/log.log",
    max_bytes: int = 10**7,
    backup_count: int = 5,
    log_level: str = "INFO",
):
    """add file handler to logger

    Args:
        file_name (str): the file name to log
        max_bytes (int, optional): max bytes of the log file. Defaults to 10**7 bytes(10MB).
        backup_count (int, optional): backup count of the log file. Defaults to 5.
        log_level (str, optional): log level. Defaults to "INFO".
    Returns:
        None
    """
    try:
        if os.path.dirname(file_name) and not os.path.isdir(os.path.dirname(file_name)):
            os.mkdir(os.path.dirname(file_name))
        file_handler = RotatingFileHandler(
            file_name, maxBytes=max_bytes, backupCount=backup_count
        )
        file_handler.setLevel(log_level_map.get(log_level, "INFO"))
        formatter = logging.Formatter(
            "%(asctime)s - dlk - %(levelname)8s - %(filename)8s:%(lineno)4d - %(message)s"
        )
        file_handler.setFormatter(formatter)
        global_logger.addHandler(file_handler)
    except Exception as e:
        global_logger.error(f"Failed to add file handler to logger: {e}")
        return



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

# TODO: WIP Comming soon
# from typing import Dict
# import torch.nn as nn
# from . import scheduler_register, scheduler_config_register, BaseScheduler
# from torch.optim.lr_scheduler import LambdaLR
# import torch.optim as optim
# import copy

# # constant

# # cos warmup
# # def lr_lambda(current_step):
# # if current_step < num_warmup_steps:
# # return float(current_step) / float(max(1, num_warmup_steps))
# # progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
# # return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))


# # constant warmup
# # def lr_lambda(current_step: int):
# # if current_step < num_warmup_steps:
# # return float(current_step) / float(max(1.0, num_warmup_steps))
# # return 1.0


# # linear warmup
# # def lr_lambda(current_step: int):
# # if current_step < num_warmup_steps:
# # return float(current_step) / float(max(1, num_warmup_steps))
# # return max(
# # 0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))


# @scheduler_config_register("constant")
# class ConstantScheduleConfig(object):
# """
# {
# config: {
# "last_epoch": -1
# },
# _name: "constant",
# }
# """
# def __init__(self, config: Dict):
# super(ConstantScheduleConfig, self).__init__()
# config = config['config']
# self.last_epoch = config["last_epoch"]


# @scheduler_register("constant")
# class ConstantSchedule(BaseScheduler):
# def __init__(self, optimizer: optim.Optimizer, config: ConstantScheduleConfig):
# super(ConstantSchedule, self).__init__()
# self.config = config
# self.optimizer = optimizer

# def get_scheduler(self):
# """TODO: Docstring for get_scheduler.
# :returns: TODO
# """
# return LambdaLR(self.optimizer, lambda _: 1, last_epoch=self.config.last_epoch)



# Copyright the author(s) of DLK.
# Copyright spark-nlp-display.
#
# This original source code is copied from spark-nlp-display and licensed under the Apache license.

import json
import os
import random
from typing import List

import numpy as np

import dlk.utils.display.style_utils as style_config
from dlk.utils.display.annotation import SpanAnnotation


class NerVisualizer:
    def __init__(self, ignore_labels: List = []):
        here = os.path.abspath(os.path.dirname(__file__))
        with open(
            os.path.join(here, "label_colors/ner.json"), "r", encoding="utf-8"
        ) as f_:
            self.label_colors = json.load(f_)
        self.ignore_labels = set(ignore_labels)

    # public function to get color for a label
    def get_label_color(self, label):
        """Returns color of a particular label

        Input: entity label <string>
        Output: Color <string> or <None> if not found
        """

        if str(label).lower() in self.label_colors:
            return self.label_colors[label.lower()]
        else:
            return None

    # private function for colors for display
    def __get_label(self, label):
        """Internal function to generate random color codes for missing colors

        Input: dictionary of entity labels and corresponding colors
        Output: color code (Hex)
        """
        if str(label).lower() in self.label_colors:
            return self.label_colors[label.lower()]
        else:
            # update it to fetch from git new labels
            r = lambda: random.randint(0, 200)
            return "#%02X%02X%02X" % (r(), r(), r())

    def set_label_colors(self, color_dict):
        """Sets label colors.
        input: dictionary of entity labels and corresponding colors
        output: self object - to allow chaining
        note: Previous values of colors will be overwritten
        """

        for key, value in color_dict.items():
            self.label_colors[key.lower()] = value
        return self

    # main display function
    def __display_ner(self, result, label_col, document_col, original_text):
        if original_text is None:
            original_text = result[document_col][0].result

        label_color = {}
        html_output = ""
        pos = 0

        sorted_labs_idx = np.argsort([int(i.begin) for i in result[label_col]])
        sorted_labs = [result[label_col][i] for i in sorted_labs_idx]

        for entity in sorted_labs:
            entity_type = entity.metadata["entity"].lower()
            if (entity_type not in label_color) and (
                entity_type not in self.ignore_labels
            ):
                label_color[entity_type] = self.__get_label(entity_type)

            begin = int(entity.begin)
            end = int(entity.end)
            if pos < begin and pos < len(original_text):
                white_text = original_text[pos:begin]
                html_output += '<span class="spark-nlp-display-others" style="background-color: white">{}</span>'.format(
                    white_text
                )
            pos = end + 1

            if entity_type in label_color:
                html_output += '<span class="spark-nlp-display-entity-wrapper" style="background-color: {}"><span class="spark-nlp-display-entity-name">{} </span><span class="spark-nlp-display-entity-type">{}</span></span>'.format(
                    label_color[entity_type],
                    original_text[begin : end + 1],  # entity.result,
                    entity.metadata["entity"],
                )
            else:
                html_output += '<span class="spark-nlp-display-others" style="background-color: white">{}</span>'.format(
                    original_text[begin : end + 1]
                )

        if pos < len(original_text):
            html_output += '<span class="spark-nlp-display-others" style="background-color: white">{}</span>'.format(
                original_text[pos:]
            )

        html_output += """</div>"""

        html_output = html_output.replace("\n", "<br>")

        return html_output

    def display(self, result):
        """Displays NER visualization.
        Inputs:
        result -- A Dataframe or dictionary.
        label_col -- Name of the column/key containing NER annotations.
        document_col -- Name of the column/key containing text document.
        original_text -- Original text of type 'str'. If specified, it will take precedence over 'document_col' and will be used as the reference text for display.
        Output: Visualization
        """
        display_result = {"entities_info": []}
        for entity in result["predict_entities_info"]:
            display_result["entities_info"].append(
                SpanAnnotation._from_dict(
                    {
                        "begin": entity["start"],
                        "end": entity["end"],
                        "metadata": {"entity": entity["labels"][0]},
                    }
                )
            )

        html_content = self.__display_ner(
            display_result,
            "entities_info",
            None,
            result["sentence"],
        )

        html_content_save = style_config.STYLE_CONFIG_ENTITIES + " " + html_content

        return html_content_save


if __name__ == "__main__":
    vis = NerVisualizer()
    result = {
        "sentence": "CRICKET - 中文字CESTERSHIRE TAKE OVER AT TOP AFTER INNINGS VICTORY .",
        "uuid": "1b764970-be58-11ec-841b-18c04d299e80",
        "entities_info": [{"start": 10, "end": 24, "labels": ["ORG"]}],
        "predict_entities_info": [
            {"start": 10, "end": 24, "labels": ["PER"]},
            {"start": 38, "end": 41, "labels": ["MISC"]},
        ],
        "predict_extend_return": {},
    }
    dis = vis.display(result)

    with open("/home/sun/Downloads/dis.html", "w") as f_:
        f_.write(dis)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, Union

import hjson
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    Parser,
    StrField,
    SubModule,
    cregister,
    init_config,
)

import dlk.data.processor
import dlk.data.subprocessor
from dlk.utils.io import open
from dlk.utils.register import register, register_module_name


class PreProcessor(object):
    """Processor"""

    def __init__(
        self,
        config: Union[str, dict],
        stage: str = "train",
        update_config: Union[Dict, None] = None,
    ):
        super(PreProcessor, self).__init__()
        config_dict = {}
        if not isinstance(config, dict):
            with open(config, "r") as f:
                config_dict = hjson.load(f, object_pairs_hook=dict)
        else:
            config_dict = config
        prepro_config = self.get_config(config_dict, update_config)
        self.init(prepro_config, stage)

    def init(self, config: Base, stage: str):
        """init the model and trainer
        Args:
            config: the preprocess config
            stage: train/predict/online

        Returns: None
        """
        # set processor
        process_name = register_module_name(config._module_name)
        processor_ins = register.get("processor", process_name)(
            config=config, stage=stage
        )
        if stage in ["train", "predict"]:
            self.processor = processor_ins.process
        else:
            assert stage == "online", f"stage {stage} is not supported"
            self.processor = processor_ins.online_process

    def get_config(self, config_dict, update_config=None):
        """get the predict config

        Args:
            config: the init config

        Returns:
            DLKPreProConfig, config_name_str
        """
        configs = Parser(config_dict, update_config=update_config).parser_init()
        assert len(configs) == 1, f"You should not use '_search' for preprocess"

        prepro_config = configs[0]["@processor"]
        return prepro_config

    def fit(self, data: Union[Dict[str, Any], pd.DataFrame]):
        """Process the data and return the processed data

        Args:
            data: {"train": train DataFrame, 'valid': valid DataFrame} if stage is not online else DataFrame

        Returns:
            processed data

        """
        return self.processor(data)



# Copyright the author(s) of DLK.

# There are many code copy from fairseq.
# Copyright (c) Facebook, Inc. and its affiliates.


import math
from typing import Dict, List, Optional

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch import Tensor

from dlk.utils.register import register

from . import Sample
from .beam_search import BeamSearch


@cregister("token_sample", "length_constrained_beam_search")
class LengthConstrainedBeamSearchConfig(Base):
    min_len_a_ratio = FloatField(
        value=0.2,
        help="the minimum length ratio, the real min_len = min_len_a_ratio * src_len + min_len_b",
    )
    min_len_b = IntField(
        value=5,
        help="the minimum length, the real min_len = min_len_a_ratio * src_len + min_len_b",
    )
    max_len_a_ratio = FloatField(
        value=2.0,
        help="the maximum length ratio, the real max_len = max_len_a_ratio * src_len + max_len_b",
    )
    max_len_b = IntField(
        value=10,
        help="the maximum length, the real max_len = max_len_a_ratio * src_len + max_len_b",
    )


@register("token_sample", "length_constrained_beam_search")
class LengthConstrainedBeamSearch(Sample):
    def __init__(self, tgt_dict, config: LengthConstrainedBeamSearchConfig):
        super().__init__(tgt_dict)
        self.min_len_a_ratio = config.min_len_a_ratio
        self.min_len_b = config.min_len_b
        self.max_len_a_ratio = config.max_len_a_ratio
        self.max_len_b = config.max_len_b
        self.beam = BeamSearch(tgt_dict)
        self.needs_src_lengths = True

    def step(
        self,
        step: int,
        lprobs,
        scores,
        prev_output_tokens: Optional[Tensor] = None,
        original_batch_idxs: Optional[Tensor] = None,
    ):
        min_lens = self.min_len_a_ratio * self.src_lengths + self.min_len_b
        max_lens = self.max_len_a_ratio * self.src_lengths + self.max_len_b
        lprobs[step < min_lens, :, self.eos] = -math.inf
        lprobs[step >= max_lens, :, self.eos] = 0
        return self.beam.step(step, lprobs, scores)



# Copyright the author(s) of DLK.

# There are many code copy from fairseq.
# Copyright (c) Facebook, Inc. and its affiliates.


from typing import Dict, List, Optional

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch import Tensor

from dlk.utils.register import register

from . import Sample


@cregister("token_sample", "beam_search")
class BeamSearchConfig(Base):
    pass


@register("token_sample", "beam_search")
class BeamSearch(Sample):
    def __init__(self, tgt_dict, config: BeamSearchConfig = None):
        super().__init__(tgt_dict)
        self.constraint_states = None

    def step(
        self,
        step: int,
        lprobs,
        scores: Optional[Tensor],
        prev_output_tokens: Optional[Tensor] = None,
        original_batch_idxs: Optional[Tensor] = None,
    ):
        bsz, beam_size, vocab_size = lprobs.size()

        if step == 0:
            # at the first step all hypotheses are equally likely, so use
            # only the first beam
            lprobs = lprobs[:, ::beam_size, :].contiguous()
        else:
            # make probs contain cumulative scores for each hypothesis
            assert scores is not None
            lprobs = lprobs + scores[:, :, step - 1].unsqueeze(-1)

        top_prediction = torch.topk(
            lprobs.view(bsz, -1),
            k=min(
                # Take the best 2 x beam_size predictions. We'll choose the first
                # beam_size of these which don't predict eos to continue with.
                beam_size * 2,
                lprobs.view(bsz, -1).size(1) - 1,  # -1 so we never select pad
            ),
        )
        scores_buf = top_prediction[0]
        indices_buf = top_prediction[1]
        # Project back into relative indices and beams
        beams_buf = torch.div(indices_buf, vocab_size, rounding_mode="trunc")
        indices_buf = indices_buf.fmod(vocab_size)

        # At this point, beams_buf and indices_buf are single-dim and contain relative indices
        return scores_buf, indices_buf, beams_buf



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import os
from typing import Callable, Dict, List, Optional, Set

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import SimpleModule
from dlk.nn.module.module_bart_encoder import BartEncoderWrap, BartEncoderWrapConfig
from dlk.utils.register import register


@cregister("encoder", "bart_like_encoder")
class BartLikeEncoderConfig(BartEncoderWrapConfig):
    """the bart encoder module"""

    class InputMap:
        encoder_attention_mask = StrField(
            value="encoder_attention_mask", help="the decoder attention mask"
        )
        encoder_head_mask = StrField(
            value="encoder_head_mask", help="the decoder head mask"
        )
        encoder_input_embedding = StrField(
            value="encoder_input_embedding", help="the input embedding"
        )

    input_map = NestField(value=InputMap, help="the input map of the linear module")
    bart_like_module_name = StrField(
        value="bart_encoder",
        help="the bart like decoder module name",
    )

    class OutputMap:
        encoder_output_embedding = StrField(
            value="encoder_output_embedding", help="the encoder outputs"
        )

    output_map = NestField(value=OutputMap, help="the output map of the linear module")


@register("encoder", "bart_like_encoder")
class BartLikeEncoder(SimpleModule):
    """Wrap for torch.nn.BartEncoder"""

    def __init__(self, config: BartLikeEncoderConfig, embedding: nn.Embedding = None):
        super(BartLikeEncoder, self).__init__(config)
        self.config = config
        self.bart_like_encoder = register.get("module", config.bart_like_module_name)(
            config=config, embedding=embedding
        )

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        self.bart_like_encoder.init_weight(method)

    def reorder_encoder_out(self, encoder_outs: Dict[str, torch.Tensor], new_order):
        """
        Reorder encoder output according to *new_order*.

        Args:
            encoder_out: output from the ``forward()`` method
            new_order (LongTensor): desired order

        Returns:
            *encoder_out* rearranged according to *new_order*
        """
        encoder_output_embedding = encoder_outs.get(
            self.config.output_map.encoder_output_embedding, None
        )
        if encoder_output_embedding is not None:
            encoder_outs[
                self.config.output_map.encoder_output_embedding
            ] = encoder_output_embedding.index_select(0, new_order)

        encoder_head_mask = encoder_outs.get(
            self.config.input_map.encoder_head_mask, None
        )
        if encoder_head_mask is not None:
            encoder_outs[
                self.config.input_map.encoder_head_mask
            ] = encoder_head_mask.index_select(0, new_order)

        encoder_input_embedding = encoder_outs.get(
            self.config.input_map.encoder_input_embedding, None
        )
        if encoder_input_embedding is not None:
            encoder_outs[
                self.config.input_map.encoder_input_embedding
            ] = encoder_input_embedding.index_select(0, new_order)

        return encoder_outs

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """All step do this

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        module_inputs = {
            "attention_mask": inputs.get(
                self.config.input_map.encoder_attention_mask, None
            ),
            "head_mask": inputs.get(self.config.input_map.encoder_head_mask, None),
            "inputs_embeds": inputs.get(
                self.config.input_map.encoder_input_embedding, None
            ),
        }

        (
            sequence_output,
            all_hidden_states,
            all_self_attentions,
        ) = self.bart_like_encoder(module_inputs)
        inputs[self.config.output_map.encoder_output_embedding] = sequence_output
        return inputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import pickle as pkl
from typing import Dict, List, Set

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import SimpleModule
from dlk.utils.io import open
from dlk.utils.register import register, register_module_name

logger = logging.getLogger(__name__)


@cregister("embedding", "static_char_cnn")
class StaticCharCNNEmbeddingConfig(Base):
    """
    the static_char_cnn embedding module config
    """

    embedding_file = StrField(
        value=MISSING,
        help="the embedding file, must be saved as numpy array by pickle",
    )
    embedding_dim = IntField(value=MISSING, minimum=0, help="the embedding dim")
    freeze = BoolField(value=False, help="whether to freeze the embedding")
    dropout = FloatField(value=0, minimum=0.0, maximum=1.0, help="dropout rate")
    padding_idx = IntField(value=0, minimum=0, help="padding index")
    kernel_sizes = ListField(
        value=[3],
        help="the kernel sizes of the cnn",
        validator=lambda x: all([i > 0 for i in x]),
    )

    class InputMap:
        char_ids = StrField(value="char_ids", help="the input name of char ids")

    input_map = NestField(
        value=InputMap,
        help="the input map of the char embedding module",
    )

    class OutputMap:
        char_embedding = StrField(
            value="char_embedding", help="the output name of char embedding"
        )

    output_map = NestField(
        value=OutputMap,
        help="the output map of the char embedding module",
    )

    submodule = SubModule(
        {
            "module": {
                "_base": "conv1d",
                "in_channels": "@$$.embedding_dim @lambda x: x",
                "out_channels": "@$$.embedding_dim @lambda x: x",
                "kernel_sizes": "@$$.kernel_sizes @lambda x: x",
            }
        },
        help="the cnn module of the char embedding",
    )


@register("embedding", "static_char_cnn")
class StaticCharCNNEmbedding(SimpleModule):
    """from 'char_ids' generate 'embedding'"""

    def __init__(self, config: StaticCharCNNEmbeddingConfig):
        super().__init__(config)
        self.config = config
        embedding_file = self.config.embedding_file
        with open(embedding_file, "rb") as f:
            embedding_file = pkl.load(f)
        cnn_configs = config._get_modules("module")
        assert len(cnn_configs) == 1
        self.cnn = register.get(
            "module", register_module_name(cnn_configs[0]._module_name)
        )(cnn_configs[0])

        self.dropout = nn.Dropout(float(self.config.dropout))
        self.embedding = nn.Embedding.from_pretrained(
            torch.tensor(embedding_file, dtype=torch.float),
            freeze=self.config.freeze,
            padding_idx=0,
        )
        assert self.embedding.weight.shape[-1] == self.config.embedding_dim

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        self.cnn.init_weight(method)
        logger.info(f"The static embedding is loaded the pretrained.")

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """fit the char embedding to cnn and pool to word_embedding

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        char_ids = inputs[self.config.input_map.char_ids]
        char_mask = (char_ids == 0).bool()
        char_embedding = self.embedding(char_ids)
        bs, seq_len, token_len, emb_dim = char_embedding.shape
        char_embedding = char_embedding.view(bs * seq_len, token_len, emb_dim)
        char_embedding = char_embedding.transpose(1, 2)
        char_embedding = self.cnn(char_embedding)  # bs*seq_len, emb_dim, token_len
        word_embedding = (
            char_embedding.masked_fill_(
                char_mask.view(bs * seq_len, 1, token_len), -1000
            )
            .max(-1)[0]
            .view(bs, seq_len, emb_dim)
            .contiguous()
        )

        inputs[self.config.output_map.char_embedding] = self.dropout(word_embedding)
        return inputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import Module


@cregister("module", "linear")
class LinearConfig:
    """
    The Linear module config
    """

    input_size = IntField(value=MISSING, minimum=0, help="the input size")
    output_size = IntField(value=MISSING, minimum=0, help="the output size")
    dropout = FloatField(value=0.0, minimum=0.0, maximum=1.0, help="the dropout rate")
    bias = BoolField(value=True, help="whether to use bias")
    pool = StrField(
        value=None,
        options=["first", None],
        help="the pooling method, currently only implement `first`",
    )


@register("module", "linear")
class Linear(Module):
    """wrap for nn.Linear"""

    def __init__(self, config: LinearConfig):
        super(Linear, self).__init__()
        self.config = config
        self.linear = nn.Linear(
            in_features=self.config.input_size, out_features=self.config.output_size
        )
        self.dropout = nn.Dropout(p=float(self.config.dropout))

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns:
            project result the shape is the same as input(no poll), otherwise depend on poll method

        """
        output = self.dropout(self.linear(input))
        if not self.config.pool:
            return output
        elif self.config.pool == "first":
            return output[:, 0]
        else:
            raise PermissionError(
                f"Currenttly we have not support the pool method '{self.config.pool}' in linear."
            )



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict

import torch.optim as optim
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch.optim.lr_scheduler import LambdaLR

from dlk.utils.register import register

from . import BaseScheduler, BaseSchedulerConfig


@cregister("scheduler", "warmup_rec_decay")
class WarmupRecDecayScheduleConfig(BaseSchedulerConfig):
    """
    the warmup_rec_decay scheduler, lr=lr*1/(1+decay)
    """

    decay = FloatField(value=0.05, help="the decay rate, lr=lr*1/(1+decay)")
    interval = StrField(
        value="epoch", options=["epoch"], help="the interval of scheduler"
    )


@register("scheduler", "rec_decay")
class WarmupRecDecaySchedule(BaseScheduler):
    """lr=lr*1/(1+decay)"""

    def __init__(
        self,
        optimizer: optim.Optimizer,
        config: WarmupRecDecayScheduleConfig,
        rt_config: Dict,
    ):
        super(WarmupRecDecaySchedule, self).__init__(optimizer, config, rt_config)
        self.config: WarmupRecDecayScheduleConfig = config

    def step_update(self, current_step: int):
        raise PermissionError("step_update is not supported in WarmupRecDecaySchedule")

    def epoch_update(self, current_epoch: int):
        if current_epoch < self.config.num_warmup_steps:
            return float(current_epoch) / float(max(1, self.config.num_warmup_steps))
        return 1 / ((1 + self.config.decay) ** current_epoch)



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from tokenizers import Tokenizer

from dlk.data.postprocessor import BasePostProcessor, BasePostProcessorConfig
from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

logger = logging.getLogger(__name__)


@cregister("postprocessor", "seq_lab")
class SeqLabPostProcessorConfig(BasePostProcessorConfig):
    """the seq_lab postprocessor"""

    use_crf = BoolField(value=False, help="whether to use crf")
    word_ready = BoolField(
        value=False, help="already gathered the first subword index of the word"
    )
    ignore_position = BoolField(
        value=False, help="whether to ignore the position of the entity"
    )
    ignore_char = ListField(
        value=[],
        suggestions=[[" ", "(", ")", "[", "]", "-", ".", ",", ":", "'", '"']],
        help="ignore the provided char if these char is prefix or suffix of the entity",
    )
    label_vocab = StrField(
        value=MISSING,
        help="the label vocab file path of entity, it should be the same as file in the preprocessor",
    )
    tokenizer_path = StrField(
        value=MISSING, help="the tokenizer file path, is not effected by `meta_dir`"
    )

    class InputMap:
        logits = StrField(value="logits", help="the output logits")
        predict_seq_label = StrField(
            value="predict_seq_label", help="the predict seq label"
        )
        index = StrField(value="_index", help="the index of the sample")

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OriginInputMap:
        uuid = StrField(value="uuid", help="the uuid or the id of the sample")
        sentence = StrField(value="sentence", help="the sentence of the sample")
        input_ids = StrField(value="input_ids", help="the input ids of the sample")
        entities_info = StrField(value="entities_info", help="the entities info")
        offsets = StrField(value="offsets", help="the offsets of the tokens")
        special_tokens_mask = StrField(
            value="special_tokens_mask", help="the special tokens mask"
        )
        word_ids = StrField(value="word_ids", help="the word ids")

    origin_input_map = NestField(
        value=OriginInputMap,
        help="the origin input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )
    aggregation_strategy = StrField(
        value="max",
        options=["none", "simple", "first", "average", "max"],
        help="the aggregation strategy for the same entity",
    )
    ignore_labels = ListField(
        value=["O", "X", "S", "E"],
        help="the ignore labels, if the entity label in this list, we will ignore this entity",
    )


class AggregationStrategy(object):
    """docstring for AggregationStrategy"""

    NONE = "none"
    SIMPLE = "simple"
    FIRST = "first"
    AVERAGE = "average"
    MAX = "max"


@register("postprocessor", "seq_lab")
class SeqLabPostProcessor(BasePostProcessor):
    """PostProcess for sequence labeling task"""

    def __init__(self, config: SeqLabPostProcessorConfig):
        super(SeqLabPostProcessor, self).__init__(config)

        self.config = config
        self.label_vocab = Vocabulary.load_from_file(
            os.path.join(self.config.meta_dir, self.config.label_vocab)
        )
        with open(self.config.tokenizer_path, "r", encoding="utf-8") as f:
            tokenizer_str = json.dumps(json.load(f))
        self.tokenizer = Tokenizer.from_str(tokenizer_str)

    def do_predict(
        self,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> List:
        """Process the model predict to human readable format

        There are three predictor for different seq_lab task dependent on the config.use_crf(the predict is already decoded to ids), and config.word_ready(subword has gathered to firstpiece)

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            all predicts

        """
        predicts = []
        if self.config.use_crf:
            predicts = self.crf_predict(
                list_batch_outputs=list_batch_outputs, origin_data=origin_data
            )
        elif self.config.word_ready:
            predicts = self.word_predict(
                list_batch_outputs=list_batch_outputs, origin_data=origin_data
            )
        else:
            predicts = self.predict(
                list_batch_outputs=list_batch_outputs, origin_data=origin_data
            )
        return predicts

    def do_calc_metrics(
        self,
        predicts: List,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            the named scores, recall, precision, f1

        """

        def _flat_entities_info(entities_info: List[Dict], text: str) -> Dict:
            """gather the same labeled entity to the same list

            Args:
                entities_info:
                    >>> [
                    >>>     {
                    >>>         "start": start1,
                    >>>         "end": end1,
                    >>>         "labels": ["label_1"]
                    >>>     },
                    >>>     {
                    >>>         "start": start2,
                    >>>         "end": end2,
                    >>>         "labels": ["label_2"]
                    >>>     },....
                    >>> ]
                text: be labeled text

            Returns:
                >>> { "label_1" [text[start1:end1]], "label_2": [text[start_2: end_2]]...}

            """
            info = {}
            for item in entities_info:
                label = item["labels"][0]
                if label not in info:
                    info[label] = []
                start_position, end_position = item["start"], item["end"]
                while start_position < end_position:
                    if text[start_position] in self.config.ignore_char:
                        start_position += 1
                    else:
                        break
                while start_position < end_position:
                    if text[end_position - 1] in self.config.ignore_char:
                        end_position -= 1
                    else:
                        break
                if (
                    start_position == end_position
                ):  # if the entity after remove ignore char be null, we set it to origin
                    start_position, end_position = item["start"], item["end"]

                if self.config.ignore_position:
                    info[label].append(text[item["start"] : item["end"]].strip())
                else:
                    info[label].append((start_position, end_position))
            return info

        all_predicts = []
        all_ground_truths = []
        for predict in predicts:
            text = predict["sentence"]
            predict_ins = _flat_entities_info(predict["predict_entities_info"], text)
            ground_truth_ins = _flat_entities_info(predict["entities_info"], text)
            all_predicts.append(predict_ins)
            all_ground_truths.append(ground_truth_ins)

        precision, recall, f1 = self.calc_score(all_predicts, all_ground_truths)
        real_name = self.loss_name_map(stage)
        logger.info(
            f"{real_name}_precision: {precision*100}, {real_name}_recall: {recall*100}, {real_name}_f1: {f1*100}"
        )
        return {
            f"{real_name}_precision": precision * 100,
            f"{real_name}_recall": recall * 100,
            f"{real_name}_f1": f1 * 100,
        }

    def calc_score(self, predict_list: List, ground_truth_list: List):
        """use predict_list and ground_truth_list to calc scores

        Args:
            predict_list: list of predict
            ground_truth_list: list of ground_truth

        Returns:
            precision, recall, f1

        """
        category_tp = {}
        category_fp = {}
        category_fn = {}

        def _care_div(a, b):
            """return a/b or 0.0 if b == 0"""
            if b == 0:
                return 0.0
            return a / b

        def _calc_num(_pred: List, _ground_truth: List):
            """calc tp, fn, fp

            Args:
                pred: pred list
                ground_truth: groud truth list

            Returns:
                tp, fn, fp

            """
            num_p = len(_pred)
            num_t = len(_ground_truth)
            truth = 0
            for p in _pred:
                if p in _ground_truth:
                    truth += 1
            return truth, num_t - truth, num_p - truth

        for predict, ground_truth in zip(predict_list, ground_truth_list):
            keys = set(list(predict.keys()) + list(ground_truth.keys()))
            for key in keys:
                tp, fn, fp = _calc_num(predict.get(key, []), ground_truth.get(key, []))
                # logger.info(f"{key} tp num {tp}, fn num {fn}, fp num {fp}")
                category_tp[key] = category_tp.get(key, 0) + tp
                category_fn[key] = category_fn.get(key, 0) + fn
                category_fp[key] = category_fp.get(key, 0) + fp

        all_tp, all_fn, all_fp = 0, 0, 0
        for key in category_tp:
            if key in self.config.ignore_labels:
                continue
            tp, fn, fp = category_tp[key], category_fn[key], category_fp[key]
            all_tp += tp
            all_fn += fn
            all_fp += fp
            precision = _care_div(tp, tp + fp)
            recall = _care_div(tp, tp + fn)
            f1 = _care_div(2 * precision * recall, precision + recall)
            logger.info(
                f"For entity 「{key}」, the precision={precision*100 :.2f}%, the recall={recall*100:.2f}%, f1={f1*100:.2f}%"
            )
        precision = _care_div(all_tp, all_tp + all_fp)
        recall = _care_div(all_tp, all_tp + all_fn)
        f1 = _care_div(2 * precision * recall, precision + recall)
        return precision, recall, f1

    def get_entity_info(
        self, sub_tokens_index: List, offset_mapping: List, word_ids: List, label: str
    ) -> Dict:
        """gather sub_tokens to get the start and end

        Args:
            sub_tokens_index: the entity tokens index list
            offset_mapping: every token offset in text
            word_ids: every token in the index of words
            label: predict label

        Returns:
            entity_info

        """
        if (
            (not sub_tokens_index)
            or (not label)
            or (label in self.config.ignore_labels)
        ):
            return {}
        start = offset_mapping[sub_tokens_index[0]][0]
        end = offset_mapping[sub_tokens_index[-1]][1]
        return {"start": start, "end": end, "labels": [label]}

    def _process4predict(
        self, predict: torch.LongTensor, index: int, origin_data: pd.DataFrame
    ) -> Dict:
        """gather the predict and origin text and ground_truth_entities_info for predict

        Args:
            predict: the predict label_ids
            index: the data index in origin_data
            origin_data: the origin pd.DataFrame

        Returns:
            >>> one_ins info
            >>> {
            >>>     "sentence": "...",
            >>>     "uuid": "..",
            >>>     "entities_info": [".."],
            >>>     "predict_entities_info": [".."],
            >>> }

        """
        one_ins = {}
        origin_ins = origin_data.iloc[int(index)]
        one_ins["sentence"] = origin_ins[self.config.origin_input_map.sentence]
        one_ins["uuid"] = origin_ins[self.config.origin_input_map.uuid]
        one_ins["entities_info"] = origin_ins[
            self.config.origin_input_map.entities_info
        ]

        word_ids = origin_ins[self.config.origin_input_map.word_ids]
        rel_token_len = len(word_ids)
        offset_mapping = origin_ins[self.config.origin_input_map.offsets][
            :rel_token_len
        ]
        predict = list(predict[:rel_token_len])
        predict_entities_info = []
        pre_label = ""
        sub_tokens_index = []
        for i, label_id in enumerate(predict):
            if offset_mapping[i] == (0, 0):  # added token like [CLS]/<s>/..
                continue
            label = self.label_vocab[label_id]
            if (
                label in self.config.ignore_labels
                or (label[0] == "B")
                or (label.split("-")[-1] != pre_label)
            ):  # label == "O" or label=='B' or label.tail != previor_label
                entity_info = self.get_entity_info(
                    sub_tokens_index, offset_mapping, word_ids, pre_label
                )
                if entity_info:
                    predict_entities_info.append(entity_info)
                pre_label = ""
                sub_tokens_index = []
            if label not in self.config.ignore_labels:
                assert len(label.split("-")) == 2
                pre_label = label.split("-")[-1]
                sub_tokens_index.append(i)
        entity_info = self.get_entity_info(
            sub_tokens_index, offset_mapping, word_ids, pre_label
        )
        if entity_info:
            predict_entities_info.append(entity_info)
        one_ins["predict_entities_info"] = predict_entities_info
        return one_ins

    def crf_predict(
        self, list_batch_outputs: List[Dict], origin_data: pd.DataFrame
    ) -> List:
        """use the crf predict label_ids get predict info

        Args:
            list_batch_outputs: the crf predict info
            origin_data: the origin data

        Returns:
            all predict instances info

        """
        if self.config.origin_input_map.sentence not in origin_data:
            logger.error(
                f"{self.config.origin_input_map.sentence} not in the origin data"
            )
            raise PermissionError(
                f"{self.config.origin_input_map.sentence} must be provided"
            )
        if self.config.origin_input_map.uuid not in origin_data:
            logger.error(f"{self.config.origin_input_map.uuid} not in the origin data")
            raise PermissionError(
                f"{self.config.origin_input_map.uuid} must be provided"
            )
        if self.config.origin_input_map.entities_info not in origin_data:
            logger.error(
                f"{self.config.origin_input_map.entities_info} not in the origin data"
            )
            raise PermissionError(
                f"{self.config.origin_input_map.entities_info} must be provided"
            )

        predicts = []
        for outputs in list_batch_outputs:
            batch_predict = outputs[self.config.input_map.predict_seq_label]
            # batch_special_tokens_mask = outputs[self.config.special_tokens_mask]

            indexes = list(outputs[self.config.input_map.index])
            outputs = []

            for predict, index in list(zip(batch_predict, indexes)):
                one_ins = self._process4predict(predict, index, origin_data)
                predicts.append(one_ins)
        return predicts

    def word_predict(
        self, list_batch_outputs: List[Dict], origin_data: pd.DataFrame
    ) -> List:
        """use the firstpiece or whole word predict label_logits get predict info

        Args:
            list_batch_outputs: the predict labels logits info
            origin_data: the origin data

        Returns:
            all predict instances info

        """
        if self.config.origin_input_map.sentence not in origin_data:
            logger.error(
                f"{self.config.origin_input_map.sentence} not in the origin data"
            )
            raise PermissionError(
                f"{self.config.origin_input_map.sentence} must be provided"
            )
        if self.config.origin_input_map.uuid not in origin_data:
            logger.error(f"{self.config.origin_input_map.uuid} not in the origin data")
            raise PermissionError(
                f"{self.config.origin_input_map.uuid} must be provided"
            )
        if self.config.origin_input_map.entities_info not in origin_data:
            logger.error(
                f"{self.config.origin_input_map.entities_info} not in the origin data"
            )
            raise PermissionError(
                f"{self.config.origin_input_map.entities_info} must be provided"
            )

        predicts = []
        for outputs in list_batch_outputs:
            batch_logits = outputs[self.config.input_map.logits].detach().cpu().numpy()

            indexes = list(outputs[self.config.input_map.index])

            outputs = []

            for logits, index in list(zip(batch_logits, indexes)):
                origin_ins = origin_data.iloc[int(index)]
                word_ids = origin_ins[self.config.origin_input_map.word_ids]

                rel_token_len = len(word_ids)
                logits = logits[:rel_token_len]

                predict = logits.argmax(-1)
                one_ins = self._process4predict(predict, index, origin_data)
                predicts.append(one_ins)
        return predicts

    def predict(
        self, list_batch_outputs: List[Dict], origin_data: pd.DataFrame
    ) -> List:
        """general predict process (especially for subword)

        Args:
            list_batch_outputs: the predict (sub-)labels logits info
            origin_data: the origin data

        Returns:
            all predict instances info

        """
        if self.config.origin_input_map.sentence not in origin_data:
            logger.error(
                f"{self.config.origin_input_map.sentence} not in the origin data"
            )
            raise PermissionError(
                f"{self.config.origin_input_map.sentence} must be provided"
            )
        if self.config.origin_input_map.uuid not in origin_data:
            logger.error(f"{self.config.origin_input_map.uuid} not in the origin data")
            raise PermissionError(
                f"{self.config.origin_input_map.uuid} must be provided"
            )
        if self.config.origin_input_map.entities_info not in origin_data:
            logger.error(
                f"{self.config.origin_input_map.entities_info} not in the origin data"
            )
            raise PermissionError(
                f"{self.config.origin_input_map.entities_info} must be provided"
            )

        predicts = []
        for outputs in list_batch_outputs:
            batch_logits = outputs[self.config.input_map.logits].detach().cpu().numpy()

            indexes = list(outputs[self.config.input_map.index])

            outputs = []

            for logits, index in list(zip(batch_logits, indexes)):
                one_ins = {}
                origin_ins = origin_data.iloc[int(index)]

                input_ids = origin_ins[self.config.origin_input_map.input_ids]
                one_ins["sentence"] = origin_ins[self.config.origin_input_map.sentence]
                one_ins["uuid"] = origin_ins[self.config.origin_input_map.uuid]
                one_ins["entities_info"] = origin_ins[
                    self.config.origin_input_map.entities_info
                ]

                rel_token_len = len(input_ids)

                special_tokens_mask = np.array(
                    origin_data.iloc[int(index)][
                        self.config.origin_input_map.special_tokens_mask
                    ][:rel_token_len]
                )
                offset_mapping = origin_data.iloc[int(index)][
                    self.config.origin_input_map.offsets
                ][:rel_token_len]

                logits = logits[:rel_token_len]

                entity_idx = logits.argmax(-1)
                labels = []
                for i, idx in enumerate(list(entity_idx)):
                    labels.append(self.label_vocab[idx])

                maxes = np.max(logits, axis=-1, keepdims=True)
                shifted_exp = np.exp(logits - maxes)
                scores = shifted_exp / shifted_exp.sum(axis=-1, keepdims=True)

                pre_entities = self.gather_pre_entities(
                    one_ins["sentence"],
                    input_ids,
                    scores,
                    offset_mapping,
                    special_tokens_mask,
                )
                grouped_entities = self.aggregate(
                    pre_entities, self.config.aggregation_strategy
                )
                # Filter anything that is in self.ignore_labels
                entities = [
                    entity
                    for entity in grouped_entities
                    if entity.get("entity", None) not in self.config.ignore_labels
                    and entity.get("entity_group", None)
                    not in self.config.ignore_labels
                ]
                predict_entities_info = []
                for entity in entities:
                    one_predict = {}
                    one_predict["start"] = entity["start"]
                    one_predict["end"] = entity["end"]
                    one_predict["labels"] = [entity["entity_group"]]
                    predict_entities_info.append(one_predict)
                one_ins["predict_entities_info"] = predict_entities_info
                predicts.append(one_ins)
        return predicts

    def aggregate(
        self, pre_entities: List[dict], aggregation_strategy: str
    ) -> List[dict]:
        if aggregation_strategy in {
            AggregationStrategy.NONE,
            AggregationStrategy.SIMPLE,
        }:
            entities = []
            for pre_entity in pre_entities:
                entity_idx = pre_entity["scores"].argmax()
                score = pre_entity["scores"][entity_idx]
                entity = {
                    "entity": self.label_vocab[entity_idx],
                    "score": score,
                    "index": pre_entity["index"],
                    "word": pre_entity["word"],
                    "start": pre_entity["start"],
                    "end": pre_entity["end"],
                }
                entities.append(entity)
        else:
            entities = self.aggregate_words(pre_entities, aggregation_strategy)

        if aggregation_strategy == AggregationStrategy.NONE:
            return entities
        return self.group_entities(entities)

    def aggregate_word(self, entities: List[dict], aggregation_strategy: str) -> dict:
        word = self.tokenizer.decode(
            [self.tokenizer.token_to_id(entity["word"]) for entity in entities]
        )
        if aggregation_strategy == AggregationStrategy.FIRST:
            scores = entities[0]["scores"]
            idx = scores.argmax()
            score = scores[idx]
            entity = self.label_vocab[idx]
        elif aggregation_strategy == AggregationStrategy.MAX:
            max_entity = max(entities, key=lambda entity: entity["scores"].max())
            scores = max_entity["scores"]
            idx = scores.argmax()
            score = scores[idx]
            entity = self.label_vocab[idx]
        elif aggregation_strategy == AggregationStrategy.AVERAGE:
            scores = np.stack([entity["scores"] for entity in entities])
            average_scores = np.nanmean(scores, axis=0)
            entity_idx = average_scores.argmax()
            entity = self.label_vocab[entity_idx]
            score = average_scores[entity_idx]
        else:
            raise ValueError("Invalid aggregation_strategy")
        new_entity = {
            "entity": entity,
            "score": score,
            "word": word,
            "start": entities[0]["start"],
            "end": entities[-1]["end"],
        }
        return new_entity

    def group_sub_entities(self, entities: List[dict]) -> dict:
        """Group together the adjacent tokens with the same entity predicted.

        Args:
            entities: The entities predicted by the pipeline.
        """
        # Get the first entity in the entity group
        entity = entities[0]["entity"].split("-")[-1]
        scores = np.nanmean([entity["score"] for entity in entities])
        tokens = [entity["word"] for entity in entities]
        entity_group = {
            "entity_group": entity,
            "score": np.mean(scores),
            "word": " ".join(tokens),
            "start": entities[0]["start"],
            "end": entities[-1]["end"],
        }
        return entity_group

    def get_tag(self, entity_name: str) -> Tuple[str, str]:
        if entity_name.startswith("B-"):
            bi = "B"
            tag = entity_name[2:]
        elif entity_name.startswith("I-"):
            bi = "I"
            tag = entity_name[2:]
        else:
            # It's not in B-, I- format
            # Default to I- for continuation.
            bi = "I"
            tag = entity_name
        return bi, tag

    def group_entities(self, entities: List[dict]) -> List[dict]:
        """Find and group together the adjacent tokens with the same entity predicted.

        Args:
            entities: The entities predicted by the pipeline.
        """

        entity_groups = []
        entity_group_disagg = []

        for entity in entities:
            if not entity_group_disagg:
                entity_group_disagg.append(entity)
                continue

            # If the current entity is similar and adjacent to the previous entity,
            # append it to the disaggregated entity group
            # The split is meant to account for the "B" and "I" prefixes
            # Shouldn't merge if both entities are B-type
            bi, tag = self.get_tag(entity["entity"])
            last_bi, last_tag = self.get_tag(entity_group_disagg[-1]["entity"])

            if tag == last_tag and bi != "B":
                # Modify subword type to be previous_type
                entity_group_disagg.append(entity)
            else:
                # If the current entity is different from the previous entity
                # aggregate the disaggregated entity group
                entity_groups.append(self.group_sub_entities(entity_group_disagg))
                entity_group_disagg = [entity]
        if entity_group_disagg:
            # it's the last entity, add it to the entity groups
            entity_groups.append(self.group_sub_entities(entity_group_disagg))

        return entity_groups

    def aggregate_words(
        self, entities: List[dict], aggregation_strategy: str
    ) -> List[dict]:
        """Override tokens from a given word that disagree to force agreement on word boundaries.

        Example:
            micro|soft| com|pany| B-ENT I-NAME I-ENT I-ENT will be rewritten with first strategy as microsoft|
            company| B-ENT I-ENT
        """
        if aggregation_strategy in {
            AggregationStrategy.NONE,
            AggregationStrategy.SIMPLE,
        }:
            raise ValueError(
                "NONE and SIMPLE strategies are invalid for word aggregation"
            )

        word_entities = []
        word_group = []
        for entity in entities:
            if not word_group:
                word_group = [entity]
            elif entity["is_subword"]:
                word_group.append(entity)
            else:
                word_entities.append(
                    self.aggregate_word(word_group, aggregation_strategy)
                )
                word_group = [entity]
        # Last item
        word_entities.append(self.aggregate_word(word_group, aggregation_strategy))
        return word_entities

    def gather_pre_entities(
        self,
        sentence: str,
        input_ids: np.ndarray,
        scores: np.ndarray,
        offset_mapping: Optional[List[Tuple[int, int]]],
        special_tokens_mask: np.ndarray,
    ) -> List[dict]:
        """Fuse various numpy arrays into dicts with all the information needed for aggregation"""
        pre_entities = []
        for idx, token_scores in enumerate(scores):
            # Filter special_tokens, they should only occur
            # at the sentence boundaries since we're not encoding pairs of
            # sentences so we don't have to keep track of those.
            if special_tokens_mask[idx]:
                continue

            word = self.tokenizer.id_to_token(int(input_ids[idx]))
            if offset_mapping is not None:
                start_ind, end_ind = offset_mapping[idx]
                word_ref = sentence[start_ind:end_ind]
                if getattr(self.tokenizer.model, "continuing_subword_prefix", None):
                    # This is a BPE, word aware tokenizer, there is a correct way
                    # to fuse tokens
                    is_subword = len(word) != len(word_ref)
                else:
                    # This is a fallback heuristic. This will fail most likely on any kind of text + punctuation mixtures that will be considered "words". Non word aware models cannot do better than this unfortunately.
                    is_subword = (
                        sentence[start_ind - 1 : start_ind] != " "
                        if start_ind > 0
                        else False
                    )

            else:
                start_ind = None
                end_ind = None
                is_subword = False

            pre_entity = {
                "word": word,
                "scores": token_scores,
                "start": start_ind,
                "end": end_ind,
                "index": idx,
                "is_subword": is_subword,
            }
            pre_entities.append(pre_entity)
        return pre_entities



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import random
import re
from typing import Dict, List

import numpy as np
import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import AdvMethod, AdvMethodConfig

logger = logging.getLogger(__name__)


@cregister("adv_method", "fgm")
class FGMAdvMethodConfig(AdvMethodConfig):
    """
    FGM adversarial training method config
    https://arxiv.org/pdf/1706.06083.pdf
    """

    embedding_pattern = StrField(
        value="model.*embedding.*embedding",
        help="the pattern of embedding name that FGM effect on",
    )
    epsilon = FloatField(
        value=1.0,
        minimum=0.0,
        help="epsilon for FGM adversarial training",
    )


@register("adv_method", "fgm")
class FGMAdvMethod(AdvMethod):
    """FGM adversarial training method"""

    def __init__(self, model: nn.Module, config: FGMAdvMethodConfig):
        super().__init__(model, config)
        self.model = model
        self.config = config
        self.backup = {}
        self.adv_para_name = set()
        for name, param in self.model.named_parameters():
            if param.requires_grad and re.findall(self.config.embedding_pattern, name):
                self.adv_para_name.add(name)
        logger.info(
            f"There are {len(self.adv_para_name)} paras will be adversarial training."
        )
        logger.info(f"They are {self.adv_para_name}.")

    def attack(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and name in self.adv_para_name:
                self.backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0:
                    r_at = self.config.epsilon * param.grad / norm
                    param.data.add_(r_at)

    def restore(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and name in self.adv_para_name:
                assert name in self.backup
                param.data = self.backup[name]
        self.backup = {}

    def training_step(self, imodel, batch: Dict[str, torch.Tensor], batch_idx: int):
        """do training_step on a mini batch

        Args:
            imodel: imodel instance
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch

        Returns:
            the outputs

        """
        optimizer = imodel.optimizers()
        rt_config = {
            "current_step": imodel.global_step,
            "current_epoch": imodel.current_epoch,
            "total_steps": imodel.num_training_steps,
            "total_epochs": imodel.num_training_epochs,
        }
        optimizer.zero_grad()
        seed = random.randint(0, int(4e9))  # 4e9 < 2e32 - 1
        torch.manual_seed(seed)  # NOTE: should fix manual seed for every forward
        np.random.seed(seed)
        result = imodel.model.training_step(batch)
        loss, _ = imodel.calc_loss(result, batch, rt_config=rt_config)
        imodel.manual_backward(loss)
        self.attack()
        result = imodel.model.training_step(batch)
        loss, loss_log = imodel.calc_loss(result, batch, rt_config=rt_config)
        imodel.manual_backward(loss)
        self.restore()
        optimizer.step()

        schedule = imodel.lr_schedulers()
        schedule.step()
        return loss, loss_log



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional

import numpy as np
import torch
import torch.nn as nn


class RoFormerSinusoidalPositionalEmbedding(nn.Embedding):
    """This module produces sinusoidal positional embeddings of any length."""

    def __init__(
        self, num_positions: int, embedding_dim: int, padding_idx: Optional[int] = None
    ) -> None:
        super().__init__(num_positions, embedding_dim)
        self.weight = self._init_weight(self.weight)

    @staticmethod
    def _init_weight(out: nn.Parameter) -> nn.Parameter:
        """
        Identical to the XLM create_sinusoidal_embeddings except features are not interleaved. The cos features are in
        the 2nd half of the vector. [dim // 2:]
        """
        n_pos, dim = out.shape
        position_enc = np.array(
            [
                [pos / np.power(10000, 2 * (j // 2) / dim) for j in range(dim)]
                for pos in range(n_pos)
            ]
        )
        out.requires_grad = False  # set early to avoid an error in pytorch-1.8+
        sentinel = dim // 2 if dim % 2 == 0 else (dim // 2) + 1
        out[:, 0:sentinel] = torch.FloatTensor(np.sin(position_enc[:, 0::2]))
        out[:, sentinel:] = torch.FloatTensor(np.cos(position_enc[:, 1::2]))
        out.detach_()
        return out

    @torch.no_grad()
    def forward(
        self, input_ids_shape: torch.Size, past_key_values_length: int = 0
    ) -> torch.Tensor:
        """`input_ids_shape` is expected to be [bsz x seqlen]."""
        bsz, seq_len = input_ids_shape[:2]
        positions = torch.arange(
            past_key_values_length,
            past_key_values_length + seq_len,
            dtype=torch.long,
            device=self.weight.device,
        )
        return super().forward(positions)

    @staticmethod
    def apply_rotary_position_embeddings(
        sinusoidal_pos, query_layer, key_layer, value_layer=None
    ):
        # https://kexue.fm/archives/8265
        # sin [batch_size, num_heads, sequence_length, embed_size_per_head//2]
        # cos [batch_size, num_heads, sequence_length, embed_size_per_head//2]
        sin, cos = sinusoidal_pos.chunk(2, dim=-1)
        # sin [θ0,θ1,θ2......θd/2-1] -> sin_pos [θ0,θ0,θ1,θ1,θ2,θ2......θd/2-1,θd/2-1]
        sin_pos = torch.stack([sin, sin], dim=-1).reshape_as(sinusoidal_pos)
        # cos [θ0,θ1,θ2......θd/2-1] -> cos_pos [θ0,θ0,θ1,θ1,θ2,θ2......θd/2-1,θd/2-1]
        cos_pos = torch.stack([cos, cos], dim=-1).reshape_as(sinusoidal_pos)
        # rotate_half_query_layer [-q1,q0,-q3,q2......,-qd-1,qd-2]
        rotate_half_query_layer = torch.stack(
            [-query_layer[..., 1::2], query_layer[..., ::2]], dim=-1
        ).reshape_as(query_layer)
        query_layer = query_layer * cos_pos + rotate_half_query_layer * sin_pos
        # rotate_half_key_layer [-k1,k0,-k3,k2......,-kd-1,kd-2]
        rotate_half_key_layer = torch.stack(
            [-key_layer[..., 1::2], key_layer[..., ::2]], dim=-1
        ).reshape_as(key_layer)
        key_layer = key_layer * cos_pos + rotate_half_key_layer * sin_pos
        if value_layer is not None:
            # rotate_half_value_layer [-v1,v0,-v3,v2......,-vd-1,vd-2]
            rotate_half_value_layer = torch.stack(
                [-value_layer[..., 1::2], value_layer[..., ::2]], dim=-1
            ).reshape_as(value_layer)
            value_layer = value_layer * cos_pos + rotate_half_value_layer * sin_pos
            return query_layer, key_layer, value_layer
        return query_layer, key_layer



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from pyecharts.charts import Gauge
from pyecharts.options import InitOpts

from dlk.display import Display, DisplayConfigBase
from dlk.utils.register import register

logger = logging.getLogger(__name__)


@cregister("display", "txt_reg")
class TextRegDisplayConfig(DisplayConfigBase):
    """the text classification display"""

    title = StrField("Text Regression Task Demo", help="the title of the demo")
    help = StrField(
        "You can add the sentence to the `sentence input area` and click `submit` to get the regression info of this sentence",
        help="the help infomation of the demo",
    )
    render_height = IntField(value=300, help="the display height")
    render_width = IntField(value=500, help="the display width")
    render_type = StrField(value="html", help="the render type of the display")
    output_head = StrField(value="Logic Regression Score", help="the output head")

    class Input:
        sentence = StrField(
            "text",
            options=["text"],
            help="the input values name for text regression, the value is the type",
        )

    input = NestField(value=Input, help="the input config")


@cregister("display", "txt_match_reg")
class TextMatchRegDisplayConfig(DisplayConfigBase):
    """the text match display"""

    title = StrField("Text Match Regression Task Demo", help="the title of the demo")
    help = StrField(
        "You can add the sentence(a and b) to the `sentence input area(a and b)` and click `submit` to get the match regression info of the pair sentences",
        help="the help infomation of the demo",
    )

    class Input:
        sentence_a = StrField("text", options=["text"], help="the text a for match")
        sentence_b = StrField("text", options=["text"], help="the text b for match")

    render_height = IntField(value=300, help="the display height")
    render_width = IntField(value=500, help="the display width")

    input = NestField(value=Input, help="the input config")


class TextRegDisplay(Display):
    """text classification display"""

    def __init__(self, config: TextRegDisplayConfig):
        super(TextRegDisplay, self).__init__(config)
        self.config = config
        self.init_options = InitOpts(
            width=f"{config.render_width}px", height=f"{config.render_height}px"
        )

    def display(self, result):
        predict = result["predict_values"][0] * 100
        return (
            Gauge(init_opts=self.init_options)
            .add(series_name="Regression", data_pair=[["", predict]])
            .render_embed()
        )


register("display", "txt_reg")(TextRegDisplay)
register("display", "txt_match_reg")(TextRegDisplay)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict

import torch.nn as nn
import torch.optim as optim
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import BaseOptimizer, BaseOptimizerConfig


@cregister("optimizer", "sgd")
class SGDOptimizerConfig(BaseOptimizerConfig):
    """the sgd optimizer"""

    lr = FloatField(value=1e-3, minimum=0.0, help="the learning rate of optimizer")
    momentum = FloatField(value=0.9, minimum=0.0, help="the momentum of sgd")
    dampening = FloatField(value=0, minimum=0, help="the dampening of sgd")
    nesterov = BoolField(value=False, help="use nesterov of sgd or not")
    weight_decay = FloatField(
        value=0.0, minimum=0.0, help="the weight decay of the optimizer"
    )


@register("optimizer", "sgd")
class SGDOptimizer(BaseOptimizer):
    """wrap for optim.SGD"""

    def __init__(self, model: nn.Module, config: SGDOptimizerConfig):
        super(SGDOptimizer, self).__init__(model, config, optim.SGD)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import importlib
import os


def import_module_dir(module_dir, namespace):
    for file in os.listdir(module_dir):
        path = os.path.join(module_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            submodule_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + submodule_name)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Any, Callable, Dict, List, Union

import hjson
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    Parser,
    StrField,
    SubModule,
    cregister,
    init_config,
)
from intc.utils import fix_trace

import dlk.adv_method
import dlk.callback
import dlk.data.data_collate
import dlk.data.datamodule
import dlk.data.dataset
import dlk.data.postprocessor
import dlk.imodel
import dlk.initmethod
import dlk.loss
import dlk.nn
import dlk.optimizer
import dlk.scheduler
import dlk.trainer
from dlk.utils.io import open
from dlk.utils.register import register, register_module_name

logger = logging.getLogger(__name__)


@cregister("fit")
class DLKFitConfig(Base):
    """"""

    specific = DictField(
        value={},
        help="""
        when we `_search` parameters we will save different config to different subdirectory. 
        it's a dict the pair is `{focus key: simplified key}`,
        `forcus key` is the path to focus key, like `@optimizer.lr`. 
        `simplified key` is the simplified name, like we can the simplify `@optimizer.lr` as `lr`. 
        if we do not provide the specific keys, we will just use 1, 2...n as the subdirectory name.
        """,
    )
    hp_metrics = StrField(
        value=None, additions=[None], help="the target metrics for logging"
    )
    log_dir = StrField(value="logs", help="the save dir of the config and logs")
    processed_data_dir = StrField(value=MISSING, help="the processed data path")
    submodule = SubModule(value={}, help="the submodule config")


class Train(object):
    """Trainer"""

    def __init__(
        self,
        config: Union[str, Dict],
        checkpoint: str = "",
        state_dict_only=True,
        strict=False,
        update_config: Union[Dict, None] = None,
    ):
        super(Train, self).__init__()
        config_dict = {}
        self.load_checkpoint_strict = strict
        if not isinstance(config, dict):
            with open(config, "r") as f:
                config_dict = hjson.load(f, object_pairs_hook=dict)
        else:
            config_dict = config

        self.checkpoint = checkpoint
        self.state_dict_only = state_dict_only
        self.configs = Parser(config_dict, update_config=update_config).parser_init()
        if self.checkpoint:
            assert (
                len(self.configs) == 1
            ), f"Reuse the checkpoint(checkpoint is not none), you must provide the (only one) config which generate the checkpoint."

        self.config_names = []
        self.hyper_parameters = []
        for i, possible_config in enumerate(self.configs):
            train_config = possible_config["@fit"]._to_dict()
            specific = train_config.get("specific", {})
            if specific:
                config_name = []
                hyper_parameter = {}
                for source, to in specific.items():
                    config_point = train_config
                    trace = fix_trace(source, train_config).split(".")
                    for t in trace:
                        config_point = config_point[t]
                    config_name.append(f"{to}={str(config_point)}")
                    hyper_parameter[to] = config_point
                self.config_names.append("/".join(config_name))
                self.hyper_parameters.append(hyper_parameter)
            else:
                self.config_names.append(str(i))
                self.hyper_parameters.append({})

    def run(self):
        """run for all configs

        Returns:
            None

        """
        logger.info(
            f"You have {len(self.config_names)} training config(s), they all will be run."
        )
        for i, (config, name, hyper_config) in enumerate(
            zip(self.configs, self.config_names, self.hyper_parameters)
        ):
            logger.info(f"Runing the {i}th {name}...")
            self.run_oneturn(config, name, hyper_config)

    def dump_config(self, config: DLKFitConfig, name: str):
        """dump the config and change the log file path to log_dir+name

        Args:
            config: the DLKFitConfig
            name: specific config name

        Returns:
            None

        """
        with open(os.path.join(config.log_dir, name, "config.json"), "w") as f:
            json.dump({"@fit": config._to_dict()}, f, ensure_ascii=False, indent=4)

    def run_oneturn(self, base_config, name, hyper_config):
        """run this config

        Args:
            base_config: Base({"@fit": '...'})
            name: config name
            hyper_config: the hyper parameters

        Returns:
            None

        """

        config: DLKFitConfig = base_config["@fit"]
        # save configure
        self.dump_config(config, name)

        # set trainer
        trainer = self.get_trainer(config, name, hyper_config)

        # register devices for valid repeat
        # set datamodule
        datamodule, data = self.get_datamodule(config, world_size=trainer.world_size)

        # init imodel and inject the origin test and valid data
        imodel = self.get_imodel(config, data, hyper_config, name)

        # start training
        if "valid" in data or "train" in data:
            trainer.fit(model=imodel, datamodule=datamodule)
        if "test" in data:
            trainer.test(model=imodel, datamodule=datamodule)

    def get_data(self, config: DLKFitConfig):
        """get the data decided by config

        Returns:
            loaded all the processed data
        """
        # NOTE: currently only support load one data for each type
        # TODO: support load multi data for each type
        data = {}
        for data_type in ["train", "valid", "test"]:
            data_path = os.path.join(config.processed_data_dir, data_type, "0.pkl")
            if os.path.exists(data_path):
                with open(data_path, "rb") as f:
                    data[data_type] = pkl.load(f)
        return data

    def get_datamodule(self, config: DLKFitConfig, world_size):
        """get the datamodule decided by config, and fit the data to datamodule

        Args:
            config: DLKFitConfig
            devices: when the devices >1 and repeat_for_valid is True, we will repeat the

        Returns:
            datamodule

        """
        data = self.get_data(config)
        datamodule_configs = config._get_modules("datamodule")
        assert len(datamodule_configs) == 1, "Currently only support one datamodule"
        data_module_config = datamodule_configs[0]
        data_module_name = register_module_name(data_module_config._module_name)
        datamodule = register.get("datamodule", data_module_name)(
            data_module_config, data, {"world_size": world_size}
        )
        return datamodule, data

    def get_trainer(self, config: DLKFitConfig, name, hyper_config):
        """get the train/predict manager decided by config

        Args:
            config: DLKFitConfig
            name: the predict progress name
            hyper_config: the hyper parameters

        Returns:
            trainer

        """
        trainer_configs = config._get_modules("trainer")
        assert len(trainer_configs) == 1, "Currently only support one trainer"
        trainer_config = trainer_configs[0]
        trainer_name = register_module_name(trainer_config._module_name)
        trainer = register.get("trainer", trainer_name)(
            trainer_config,
            rt_config={
                "log_dir": config.log_dir,
                "name": name,
                "hp_metrics": config.hp_metrics,
                "hyper_config": hyper_config,
            },
        )
        return trainer

    def get_imodel(self, config: DLKFitConfig, data, hyper_config, name):
        """get the imodel decided by config, and inject the origin test and valid data

        Args:
            config: DLKFitConfig
            data: {"train": '..', 'valid': '..', ..}

        Returns:
            imodel

        """
        imodel_configs = config._get_modules("imodel")
        assert len(imodel_configs) == 1, "Currently only support one imodel"
        imodel_config = imodel_configs[0]
        imodel_name = register_module_name(imodel_config._module_name)
        imodel = register.get("imodel", imodel_name)(
            imodel_config,
            rt_config={
                "hp_metrics": config.hp_metrics,
                "hyper_config": hyper_config,
                "name": name,
                "log_dir": config.log_dir,
            },
        )
        if self.checkpoint:
            logger.info(f"reuse the checkpoint at {self.checkpoint}")
            if self.state_dict_only:
                with open(self.checkpoint, mode="rb") as f:
                    state_dict = torch.load(self.checkpoint)["state_dict"]
                imodel.load_state_dict(state_dict, strict=self.load_checkpoint_strict)
            else:
                raise NotImplementedError(
                    f"Currently not implement the load_from_checkpoint, only support load the `state_dict`"
                )
                # imodel.load_from_checkpoint(self.checkpoint)
        if "valid" in data:
            imodel._origin_data["valid"] = data["valid"]

        if "test" in data:
            imodel._origin_data["test"] = data["test"]

        return imodel



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
from functools import partial
from typing import Callable, Dict, List, Set

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "token2charid")
class Token2CharIDConfig(BaseSubProcessorConfig):
    """the token 2 character id subprocessor"""

    train_data_set = ListField(
        value=["train", "valid", "test"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )
    predict = ListField(
        value=["predict"],
        suggestions=[["predict"]],
        help="the data set should be processed for predict stage",
    )
    online = ListField(
        value=["online"],
        suggestions=[["online"]],
        help="the data set should be processed for online stage",
    )

    class InputMap:
        sentence = StrField(value="sentence", help="the sentence")
        offsets = StrField(value="offsets", help="the offsets")

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OutputMap:
        char_ids = StrField(value="char_ids", help="the char ids")

    output_map = NestField(
        value=OutputMap,
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )
    vocab = StrField(value="char_vocab.json", help="the vocab file for the character")
    max_token_len = IntField(
        value=20,
        minimum=1,
        help="the max length of token, then the output will be max_token_len x token_num (put max_token_len in previor is for padding on token_num)",
    )


@register("subprocessor", "token2charid")
class Token2CharID(BaseSubProcessor):
    """Use 'Vocabulary' map the character from tokens to id"""

    def __init__(self, stage: str, config: Token2CharIDConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.stage = stage
        self.config = config
        self.vocab: Vocabulary = None

    def load_meta(self):
        self.loaded_meta = True
        self.vocab = Vocabulary.load_from_file(
            os.path.join(self.meta_dir, self.config.vocab)
        )

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """firstpiece relabel the data
        one_token like 'apple' will generate [1, 2, 2, 3] if max_token_len==4 and the vocab.word2idx = {'a': 1, "p": 2, "l": 3}

        Args:
            data: will processed data

            deliver_meta:
                ignore
        Returns:
            relabeld data
        """
        if not self.loaded_meta:
            self.load_meta()

        def get_index_wrap(sentence_name, offset_name, x):
            """wrap the vocab.get_index"""
            sentence = list(x[sentence_name])
            offsets = x[offset_name]
            char_ids = []
            for offset in offsets:
                token = sentence[offset[0] : offset[1]][: self.config.max_token_len]
                token = token + [self.vocab.pad] * (
                    self.config.max_token_len - len(token)
                )
                char_ids.append([self.vocab.get_index(c) for c in token])
            return char_ids

        get_index = partial(
            get_index_wrap,
            self.config.input_map.sentence,
            self.config.input_map.offsets,
        )
        data[self.config.output_map.char_ids] = data.apply(get_index, axis=1)
        return data



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import abc
import logging
from typing import Any, Callable, Dict, List, Set

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register, register_module_name

logger = logging.getLogger(__name__)


class ModuleOutputRenameMixin:
    """Just rename the output key name by config to adapt the input field of downstream module."""

    def dict_rename(self, input: Dict, output_map: Dict[str, str]) -> Dict:
        """rename the key of input(dict) by output_map(name map)

        Args:
            input: will rename input
            output_map:  name map

        Returns:
            renamed input
        """
        if isinstance(input, dict):
            output = {}
            for key, value in input.items():
                if key in output_map:
                    output[output_map[key]] = value
                else:
                    output[key] = value
            return output
        else:
            raise PermissionError("Not Defined")

    def get_real_name(self, name: str, name_map: Dict[str, str]) -> str:
        """use the name_map to map the input name to real name

        Args:
            name: input_name
            name_map: name map
        Returns:
            real_name

        """
        if name in name_map:
            return name_map[name]
        else:
            return name

    def get_input_name(self, name: str) -> str:
        """use config._input_map map the name to real name

        Args:
            name: input_name

        Returns:
            real_name

        """
        return self.get_real_name(name, self.__input_map)

    def get_output_name(self, name: str) -> str:
        """use config._output_map map the name to real name

        Args:
            name: output_name

        Returns:
            real_name

        """
        return self.get_real_name(name, self.__output_map)

    def set_rename(self, input: Set, output_map: Dict[str, str]) -> Set:
        """rename all the name in input by output_map

        Args:
            input: a set of names
            output_map: name map

        Returns:
            renamed input

        """
        if isinstance(input, set):
            output = set()
            for key in input:
                if key in output_map:
                    output.add(output_map[key])
                else:
                    output.add(key)
            return output
        else:
            raise PermissionError("Not Defined")


class GatherOutputMixin(object):
    """gather all the small batches output to a big batch"""

    @staticmethod
    def proc_dist_outputs(dist_outputs: List[Dict]) -> List[Dict]:
        """gather all distributed outputs to outputs which is like in a single worker.

        Args:
            dist_outputs: the inputs of pytorch_lightning train/test/.._epoch_end when using ddp
        Returns:
            the inputs of pytorch_lightning train/test/.._epoch_end when only run on one worker.

        """
        outputs = []
        for dist_output in dist_outputs:
            one_output = {}
            for key in dist_output:
                try:
                    one_output[key] = torch.cat(
                        torch.swapaxes(dist_output[key], 0, 1).unbind(), dim=0
                    )
                except:
                    raise KeyError(f"{key}: {dist_output[key]}")
            outputs.append(one_output)
        return outputs

    def gather_outputs(self, outputs: List[Dict]):
        """gather the dist outputs

        Args:
            outputs: one node outputs

        Returns:
            all outputs

        """
        if self.trainer.world_size > 1:
            dist_outputs = self.all_gather(
                outputs
            )  # WARN: must padding all the dim to same except batch_size, otherwise will be trunc by the default all_gather function. Or check pytorch_lightning/utilities/distributed.gather_all_tensors
            if self.local_rank in [0, -1]:
                outputs = self.proc_dist_outputs(dist_outputs)
        return outputs

    def concat_list_of_dict_outputs(self, outputs: List[Dict]) -> Dict:
        """only support all the outputs has the same dim, now is deprecated.

        Args:
            outputs: multi node returned output (list of dict)

        Returns:
            Concat all list by name

        """
        key_all_batch_map = {}
        for batch in outputs:
            for key in batch:
                if key not in key_all_batch_map:
                    key_all_batch_map[key] = []
                key_all_batch_map[key].append(batch[key])
        key_all_ins_map = {}
        for key in key_all_batch_map:
            key_all_ins_map[key] = torch.cat(key_all_batch_map[key], dim=0)

        return key_all_ins_map


class IModuleIO(metaclass=abc.ABCMeta):
    """Currently Deprecated:
    interface for check the modules input and output"""

    @abc.abstractmethod
    def provide_keys(self) -> List[str]:
        """return all keys of the dict of the module returned

        Returns:
            all keys
        """
        pass

    @abc.abstractmethod
    def check_keys_are_provided(self, provide: List[str]) -> bool:
        """check this module required key are provided

        Returns:
            pass or not

        """
        pass

    def check_module_chain(self, module_list: List["BaseModule"]) -> bool:
        """check the interfaces of the list of modules are aligned or not.

        Args:
            module_list: a series modules

        Returns:
            pass or not

        Raises:
            ValueError: the check is not passed
        """
        assert len(module_list) > 1
        result = True
        for i in range(len(module_list) - 1):
            result = result and module_list[i + 1].check_keys_are_provided(
                module_list[i].provide_keys()
            )
            if not result:
                raise ValueError(
                    f'The module "{module_list[i+1]._name}" is required "{", ".join(module_list[i+1].provide_keys())}", \
                    but the module "{module_list[i]._name}" provide "{", ".join(module_list[i].provide_keys())}"! '
                )
        return True


class IModuleStep(metaclass=abc.ABCMeta):
    """docstring for ModuleStepMixin"""

    @abc.abstractmethod
    def predict_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do predict for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            the predicts outputs

        """
        raise NotImplementedError

    @abc.abstractmethod
    def training_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do training for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        raise NotImplementedError

    @abc.abstractmethod
    def validation_step(
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """do validation for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        raise NotImplementedError

    def test_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do test for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        return self.validation_step(inputs)


class BaseModel(nn.Module, IModuleStep):
    """All pytorch models should inheritance this class"""

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """all models should apply this method

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        raise NotImplementedError


class BaseModule(nn.Module, IModuleStep):
    """All pytorch modules should inheritance this class"""

    def __init__(self, config: Base):
        super(BaseModule, self).__init__()
        dict_config = config._to_dict()
        self.__logits_gather = None
        for key, config in config._get_named_modules("module").items():
            if register_module_name(config._module_name) == "logits_gather":
                self.__logits_gather = register("module", "logits_gather")(config)

    def reorder_encoder_out(self, encoder_outs: Dict[str, torch.Tensor], new_order):
        """
        Reorder encoder output according to *new_order*.

        Args:
            encoder_out: output from the ``forward()`` method
            new_order (LongTensor): desired order

        Returns:
            *encoder_out* rearranged according to *new_order*
        """
        raise NotImplementedError

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        for module in self.children():
            module.apply(method)

    def forward_logits_gather(
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """all module should apply this method

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        inputs = self.forward(inputs)
        if self.__logits_gather is not None:
            inputs = self.__logits_gather(inputs)
        return inputs

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """in simple module, all step fit to this method

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        raise NotImplementedError


class SimpleModule(BaseModule):
    """SSimpleModule, all train/predict/test/validation step call the forward"""

    def predict_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do predict for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        return self(inputs)

    def training_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do train for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        return self(inputs)

    def validation_step(
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """do validation for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        return self(inputs)

    def test_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do test for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        return self(inputs)


class BaseIdentityModule(SimpleModule):
    """docstring for IdentityModule"""

    def __init__(self, config=Base()):
        super(BaseIdentityModule, self).__init__(config)

    def init_weight(self, method):
        pass

    def forward(self, inputs):
        return inputs

    def forward_wrap(self, inputs):
        return inputs



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
from functools import partial
from typing import Callable, Dict, Iterable, List, Set, Union

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from tokenizers import Tokenizer
from transformers import ViTImageProcessor

from dlk.data.subprocessor import BaseSubProcessor, BaseSubProcessorConfig
from dlk.utils.io import open
from dlk.utils.register import register

logger = logging.getLogger(__name__)

hf_image_processor = {"vit": ViTImageProcessor}


@cregister("subprocessor", "image_process")
class ImageProcessConfig(BaseSubProcessorConfig):
    """the token norm subprocessor"""

    train_data_set = ListField(
        value=["train", "valid", "test"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )
    predict_data_set = ListField(
        value=["predict"],
        suggestions=[["predict"]],
        help="the data set should be processed for predict stage",
    )
    online_data_set = ListField(
        value=["online"],
        suggestions=[["online"]],
        help="the data set should be processed for online stage",
    )

    preprocess_config = StrField(
        value=MISSING,
        suggestions=["preprocess_config.json"],
        help="the hf image preprocess config path",
    )
    preprocess_method = StrField(
        value="vit",
        options=list(hf_image_processor.keys()),
        help="the hf image preprocess method",
    )

    class InputMap:
        image = StrField(
            value="image",
            suggestions=["image"],
            help="the input image",
        )

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OutputMap:
        pixel_values = StrField(
            value="pixel_values",
            suggestions=["pixel_values"],
            help="the processed image values",
        )

    output_map = NestField(
        value=OutputMap,
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )


@register("subprocessor", "image_process")
class ImageProcess(BaseSubProcessor):
    """preprocess image"""

    def __init__(self, stage: str, config: ImageProcessConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.stage = stage
        self.config = config

        self.image_processor = hf_image_processor[
            self.config.preprocess_method
        ].from_json_file(self.config.preprocess_config)

    def image_process(self, input: pd.Series) -> np.ndarray:
        """norm token, the result len(result) == len(token), exp.  12348->00000

        Args:
            input:  include image

        Returns:
            normed_token

        """
        return self.image_processor(
            input[self.config.input_map.image], return_tensors="np"
        ).pixel_values

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """image process entry

        Args:
            data:
            >>> |image                                           |
            >>> |------------------------------------------------|
            >>> |PIL.JpegImagePlugin.JpegImageFile image mode=...|
            >>> |PIL.JpegImagePlugin.JpegImageFile image mode=...|

            deliver_meta:
                False
        Returns:
            processed data

        """
        data[self.config.output_map.pixel_values] = data.apply(
            self.image_process, axis=1
        )

        return data



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import copy
import json
import logging
import os
import pickle as pkl
import uuid
from typing import Any, Callable, Dict, List, Union

import hjson
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    Parser,
    StrField,
    SubModule,
    cregister,
    init_config,
)

from dlk.train import DLKFitConfig
from dlk.utils.io import open
from dlk.utils.register import register, register_module_name

logger = logging.getLogger(__name__)


class Predict(object):
    """Predict

    Config Example:
    """

    def __init__(
        self,
        config: Union[str, dict],
        checkpoint: str,
        update_config: Union[Dict, None] = None,
    ):
        super(Predict, self).__init__()
        config_dict = {}
        self.online = False
        if not isinstance(config, dict):
            with open(config, "r") as f:
                config_dict = hjson.load(f, object_pairs_hook=dict)
        else:
            config_dict = config
        if isinstance(checkpoint, str):
            with open(checkpoint, "rb") as f:
                self.checkpoint = torch.load(f, map_location=torch.device("cpu"))
        else:
            self.checkpoint = torch.load(checkpoint, map_location=torch.device("cpu"))
        dlk_config, name_str = self.get_config(config_dict, update_config)
        self.dlk_config = dlk_config
        self.init(dlk_config, name_str)

    def get_config(self, config_dict, update_config):
        """get the predict config

        Args:
            config: the init config

        Returns:
            DLKFitConfig, config_name_str
        """
        configs = Parser(config_dict, update_config=update_config).parser_init()
        assert len(configs) == 1, f"You should not use '_search' for predict/online"

        fit_config: DLKFitConfig = configs[0]["@fit"]

        config_name_str = "predict"
        return fit_config, config_name_str

    def init(self, config, name):
        """init the model and trainer
        Args:
            config: the config
            name: the name of the config

        Returns: None
        """
        # set trainer
        self.trainer = self.get_trainer(config, name)

        # init imodel and inject the origin test and valid data
        self.imodel = self.get_imodel(config)

    def predict(self, data=None, save_condition=False):
        """init the model, datamodule, manager then predict the predict_dataloader

        Args:
            data: if provide will not load from data_path

        Returns:
            None

        """
        # set datamodule
        datamodule, data = self.get_datamodule(
            self.dlk_config, data, world_size=self.trainer.world_size
        )

        # start predict
        with torch.no_grad():
            predict_result = self.trainer.predict(
                model=self.imodel, datamodule=datamodule
            )
        return self.imodel.postprocessor(
            stage="predict",
            list_batch_outputs=predict_result,
            origin_data=data["predict"],
            rt_config={},
            save_condition=save_condition,
        )

    def get_data(self, config):
        """get the data decided by config

        Args:
            config: {"config": {"data_path": '..'}}

        Returns:
            loaded data

        """
        data = {}
        for data_type in ["predict"]:
            data_path = os.path.join(config.processed_data_dir, data_type, "0.pkl")
            if os.path.exists(data_path):
                with open(data_path, "rb") as f:
                    data[data_type] = pkl.load(f)
        return data

    def get_datamodule(self, config, data, world_size):
        """get the datamodule decided by config, and fit the data to datamodule

        Args:
            config: {"task": {"datamodule": '..'}}
            data: {"train": '..', 'valid': '..', ..}

        Returns:
            datamodule

        """
        if not data and not self.online:
            data = self.get_data(config)
        datamodule_configs = config._get_modules("datamodule")
        assert len(datamodule_configs) == 1, "Currently only support one datamodule"
        data_module_config = datamodule_configs[0]
        data_module_name = register_module_name(data_module_config._module_name)
        datamodule = register.get("datamodule", data_module_name)(
            data_module_config, data, {"world_size": world_size}
        )
        return datamodule, data

    def get_trainer(self, config: DLKFitConfig, name):
        """get the train/predict manager decided by config

        Args:
            config: DLKFitConfig
            name: the predict progress name

        Returns:
            trainer

        """
        trainer_configs = config._get_modules("trainer")
        assert len(trainer_configs) == 1, "Currently only support one trainer"
        trainer_config = trainer_configs[0]
        trainer_name = register_module_name(trainer_config._module_name)
        trainer = register.get("trainer", trainer_name)(
            trainer_config, rt_config={"log_dir": config.log_dir, "name": name}
        )
        return trainer

    def get_imodel(self, config):
        """get the imodel decided by config

        Args:

        Returns:
            imodel

        """
        imodel_configs = config._get_modules("imodel")
        assert (
            len(imodel_configs) == 1
        ), f"Currently only support one imodel, {imodel_configs}"
        imodel_config = imodel_configs[0]
        imodel_name = register_module_name(imodel_config._module_name)
        imodel = register.get("imodel", imodel_name)(imodel_config, checkpoint=True)
        if self.checkpoint:
            imodel.load_state_dict(self.checkpoint["state_dict"], strict=True)
        imodel.eval()
        return imodel



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import pickle as pkl
from typing import Callable, Dict, List, Set

import numpy as np
import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import SimpleModule
from dlk.utils.register import register


@cregister("embedding", "random")
class RandomEmbeddingConfig(Base):
    """
    the random embedding module config
    """

    vocab_size = IntField(
        value=0,
        minimum=0,
        help="the vocab size, when the vocab size is 0, it means this class is overload by other Embedding(like Static)",
    )
    embedding_dim = IntField(value=MISSING, minimum=1, help="the embedding dim")
    dropout = FloatField(value=0, minimum=0.0, maximum=1.0, help="dropout rate")
    padding_idx = IntField(value=0, minimum=0, help="padding index")

    class OutputMap:
        embedding = StrField(value="embedding", help="the output of embedding name")

    output_map = NestField(
        value=OutputMap, help="the output map of the random embedding module"
    )

    class InputMap:
        input_ids = StrField(value="input_ids", help="the input of input_ids name")

    input_map = NestField(
        value=InputMap, help="the input map of the random embedding module"
    )


@register("embedding", "random")
class RandomEmbedding(SimpleModule):
    """from 'input_ids' generate 'embedding'"""

    def __init__(self, config: RandomEmbeddingConfig):
        super().__init__(config)
        self.config = config
        self.dropout = nn.Dropout(float(self.config.dropout))
        normal = torch.distributions.Normal(
            torch.tensor([0.0]), torch.tensor([2.0 / self.config.embedding_dim])
        )
        if self.config.vocab_size:
            self.embedding = nn.Embedding.from_pretrained(
                normal.sample(
                    (self.config.vocab_size, self.config.embedding_dim)
                ).squeeze_(-1),
                padding_idx=self.config.padding_idx,
            )

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        self.embedding.apply(method)

    def share_embedding(self, embedding):
        """link the embedding.embedding to self.embedding

        Args:
            embedding: source embedding

        Returns:
            None

        """
        self.embedding = embedding.embedding

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """get the random embedding

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        inputs[self.config.output_map.embedding] = self.dropout(
            self.embedding(inputs[self.config.input_map.input_ids])
        )
        return inputs



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import random
import re
from typing import Dict, List

import numpy as np
import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import AdvMethod, AdvMethodConfig

logger = logging.getLogger(__name__)


@cregister("adv_method", "pgd")
class PGDAdvMethodConfig(AdvMethodConfig):
    """
    PGD adversarial training method
    """

    embedding_pattern = StrField(
        value="model.*embedding.*embedding",
        help="the pattern of embedding name that PGD effect on",
    )
    epsilon = FloatField(
        value=1.0,
        minimum=0.0,
        help="epsilon for PGD adversarial training",
    )
    alpha = FloatField(
        value=0.3,
        minimum=0.0,
        help="alpha for PGD adversarial training",
    )
    adv_k = IntField(value=3, minimum=1, help="PGD adversarial training times")


@register("adv_method", "pgd")
class PGDAdvMethod(AdvMethod):
    """PGD adversarial training method"""

    def __init__(self, model: nn.Module, config: PGDAdvMethodConfig):
        super().__init__(model, config)
        self.model = model
        self.config = config
        self.emb_backup = {}
        self.grad_backup = {}
        self.adv_para_name = set()
        for name, param in self.model.named_parameters():
            if param.requires_grad and re.findall(self.config.embedding_pattern, name):
                self.adv_para_name.add(name)
        logger.info(
            f"There are {len(self.adv_para_name)} paras will be adversarial training."
        )
        logger.info(f"They are {self.adv_para_name}.")

    def attack(self, is_first_attack=False):
        for name, param in self.model.named_parameters():
            if param.requires_grad and name in self.adv_para_name:
                if is_first_attack:
                    self.emb_backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0 and not torch.isnan(norm):
                    r_at = self.config.alpha * param.grad / norm
                    param.data.add_(r_at)
                    param.data = self.project(name, param.data, self.config.epsilon)

    def project(self, param_name, param_data, epsilon):
        r = param_data - self.emb_backup[param_name]
        if torch.norm(r) > epsilon:
            r = epsilon * r / torch.norm(r)
        return self.emb_backup[param_name] + r

    def restore(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and name in self.adv_para_name:
                assert name in self.emb_backup
                param.data = self.emb_backup[name]
        self.emb_backup = {}

    def backup_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                self.grad_backup[name] = param.grad.clone()

    def restore_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                param.grad = self.grad_backup[name]

    def training_step(self, imodel, batch: Dict[str, torch.Tensor], batch_idx: int):
        """do training_step on a mini batch

        Args:
            imodel: imodel instance
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch

        Returns:
            the outputs

        """
        optimizer = imodel.optimizers()
        rt_config = {
            "current_step": imodel.global_step,
            "current_epoch": imodel.current_epoch,
            "total_steps": imodel.num_training_steps,
            "total_epochs": imodel.num_training_epochs,
        }
        optimizer.zero_grad()
        seed = random.randint(0, int(4e9))  # 4e9 < 2e32 - 1
        torch.manual_seed(seed)  # NOTE: should fix manual seed for every forward
        np.random.seed(seed)
        result = imodel.model.training_step(batch)
        loss, loss_log = imodel.calc_loss(result, batch, rt_config=rt_config)
        imodel.manual_backward(loss)

        self.backup_grad()
        for t in range(self.config.adv_k):
            self.attack(is_first_attack=(t == 0))

            if t != self.config.adv_k - 1:
                optimizer.zero_grad()
            else:
                self.restore_grad()
            result = imodel.model.training_step(batch)
            loss, loss_log = imodel.calc_loss(result, batch, rt_config=rt_config)
            imodel.manual_backward(loss)
        self.restore()
        optimizer.step()

        schedule = imodel.lr_schedulers()
        schedule.step()
        return loss, loss_log



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import BaseLoss, BaseLossConfig


@cregister("loss", "identity")
class IdentityLossConfig(BaseLossConfig):
    """do not calc loss in this step, but calculated in previous"""

    loss = StrField(value="loss", help="the pre calculated loss result")


@register("loss", "identity")
class IdentityLoss(BaseLoss):
    """do not calc loss in this step, but calculated in previous"""

    def __init__(self, config: IdentityLossConfig):
        super(IdentityLoss, self).__init__(config)
        self.config: IdentityLossConfig

    def _calc(self, result, inputs, rt_config, scale):
        """calc the loss the predict is from result, the ground truth is from inputs

        Args:
            result: the model predict dict
            inputs: the all inputs for model
            rt_config: provide the current training status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            scale: the scale rate for current stage
        Returns:
            loss

        """

        if rt_config["current_step"] > self.config.schedule[self.current_stage]:
            self.current_stage += 1
        scale = self.config.scale[self.current_stage]
        loss = result[self.config.loss] * scale
        return loss, {self.config.log_map.loss: loss}



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Dict

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch.nn.utils.rnn import pad_sequence

from dlk.utils.register import register


@cregister("data_collate", "default")
class DefaultCollateConfig:
    """default collate function"""

    key_padding_pairs = DictField(
        value={}, help="the pair of key and padding value, the data is 1d"
    )
    key_padding_pairs_2d = DictField(
        value={}, help="the pair of key and padding value, the data is 2d"
    )
    key_no_padding = ListField(value=[], help="the key just concat no padding")
    key_padding_pairs_3d = DictField(
        value={}, help="the pair of key and padding value, the data is 3d"
    )
    gen_mask = DictField(value={}, help="the pair of key and generated mask key")


@register("data_collate", "default")
class DefaultCollate(object):
    """default collate function"""

    def __init__(self, config: DefaultCollateConfig):
        super(DefaultCollate, self).__init__()
        self.config = config

    def __call__(self, batch):
        keys = batch[0].keys()
        data_map: Dict[str, Any] = {}
        for key in keys:
            data_map[key] = []
        for key in keys:
            for one_ins in batch:
                data_map[key].append(one_ins[key])
        if self.config.gen_mask:
            for key, mask in self.config.gen_mask.items():
                if key not in data_map:
                    continue
                data_map[mask] = []
                for item in data_map[key]:
                    data_map[mask].append(
                        torch.tensor([1] * len(item), dtype=torch.int)
                    )
        for key in data_map:
            if key in self.config.key_no_padding:
                data = [ins for ins in data_map[key]]
                data_map[key] = torch.cat(data, dim=0)
            elif key in self.config.key_padding_pairs_3d:
                max_x, max_y, max_z = 0, 0, 0
                for ins in data_map[key]:
                    cur_x, cur_y, cur_z = ins.shape
                    max_x = max(max_x, cur_x)
                    max_y = max(max_y, cur_y)
                    max_z = max(max_z, cur_z)
                _data = torch.full(
                    (len(data_map[key]), max_x, max_y, max_z),
                    fill_value=self.config.key_padding_pairs_3d[key],
                    dtype=data_map[key][0].dtype,
                )
                for i, ins in enumerate(data_map[key]):
                    cur_x, cur_y, cur_z = ins.shape
                    _data[i][:cur_x, :cur_y, :cur_z] = ins
                data_map[key] = _data
            elif key in self.config.key_padding_pairs_2d:
                max_m, max_n = 0, 0
                for ins in data_map[key]:
                    cur_m, cur_n = ins.shape
                    max_m = max(max_m, cur_m)
                    max_n = max(max_n, cur_n)
                _data = torch.full(
                    (len(data_map[key]), max_m, max_n),
                    fill_value=self.config.key_padding_pairs_2d[key],
                    dtype=data_map[key][0].dtype,
                )
                for i, ins in enumerate(data_map[key]):
                    cur_m, cur_n = ins.shape
                    _data[i][:cur_m, :cur_n] = ins
                data_map[key] = _data
            elif key == "_index":
                _data = pad_sequence(
                    [i.unsqueeze(0) for i in data_map[key]],
                    batch_first=True,
                    padding_value=0,
                ).squeeze()
                if not _data.size():
                    _data.unsqueeze_(0)
                data_map[key] = _data
            else:
                try:
                    data_map[key] = pad_sequence(
                        data_map[key],
                        batch_first=True,
                        padding_value=self.config.key_padding_pairs.get(key, 0),
                    )
                except:
                    # if the data_map[key] is size 0, we can concat them
                    if data_map[key][0].size():
                        raise ValueError(
                            f"The {data_map[key]} can not be concat by pad_sequence."
                        )
                    _data = pad_sequence(
                        [i.unsqueeze(0) for i in data_map[key]],
                        batch_first=True,
                        padding_value=self.config.key_padding_pairs[key],
                    ).squeeze()
                    if not _data.size():
                        _data.unsqueeze_(0)
                    data_map[key] = _data
        return data_map



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.display import Display, DisplayConfigBase
from dlk.utils.display.relation_extraction import RelationExtractionVisualizer
from dlk.utils.register import register

logger = logging.getLogger(__name__)


@cregister("display", "span_relation")
class RelationDisplayConfig(DisplayConfigBase):
    """the relation extraction display"""

    title = StrField("Relation Extraction Task Demo", help="the title of the demo")
    help = StrField(
        "You can add the sentence to the `sentence input area` and click `submit` to get the relations info of this sentence",
        help="the help infomation of the demo",
    )
    render_height = IntField(value=500, help="the display height")
    render_width = IntField(value=900, help="the display width")
    render_type = StrField(value="html", help="the render type of the display")

    output_head = StrField(value="Relations", help="the output head")

    class Input:
        sentence = StrField(
            "text",
            options=["text"],
            help="the input values name of relations extraction, the value is the type",
        )

    input = NestField(value=Input, help="the input config")

    class Hold:
        entities_info = ListField([], help="the value of entities_info")
        relations_info = ListField([], help="the value of relations_info")

    hold = NestField(value=Hold, help="the hold values")

    ignore_relations = ListField(
        value=["O"],
        help="the relation types which is ignored for display",
    )


@register("display", "span_relation")
class RelationDisplay(Display):
    """relation extraction display"""

    def __init__(self, config: RelationDisplayConfig):
        super(RelationDisplay, self).__init__(config)
        self.visual = RelationExtractionVisualizer(
            config.ignore_relations, width=config.render_width
        )

    def display(self, result):
        return self.visual.display(result)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import BaseLoss, BaseLossConfig


@cregister("loss", "mse")
class MSELossConfig(BaseLossConfig):
    """the MEAN SQUARE ERROR loss(MSE AKA L2 loss)"""

    pass


@register("loss", "mse")
class MSELoss(BaseLoss):
    """mse loss for regression, distill, etc."""

    def __init__(self, config: MSELossConfig):
        super(MSELoss, self).__init__(config)
        self.config: MSELossConfig
        self.mse = nn.MSELoss(
            reduction=self.config.reduction
        )  # we will use masked_select, and apply reduction=batchmean (sum/batch_size)

    def _calc(self, result, inputs, rt_config, scale):
        """calc the loss the predict is from result, the ground truth is from inputs

        Args:
            result: the model predict dict
            inputs: the all inputs for model
            rt_config: provide the current training status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            scale: the scale rate for current stage
        Returns:
            loss

        """
        pred = result[self.pred_name]
        target = inputs[self.truth_name]
        batch_size = target.shape[0]
        loss = self.mse(pred, target) * scale / batch_size  # batch mean
        return loss, {self.config.log_map.loss: loss}



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Set

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import SimpleModule
from dlk.nn.module.bert_like import BertLike, BertLikeConfig
from dlk.utils.register import register


@cregister("embedding", "bert_like")
class BertLikeEmbeddingConfig(BertLikeConfig):
    """
    the bert like embedding module
    """

    embedding_dim = IntField(value=MISSING, minimum=1, help="the embedding dim")

    class InputMap:
        input_ids = StrField(value="input_ids", help="the input ids")
        attention_mask = StrField(value="attention_mask", help="the attention mask")
        type_ids = StrField(value="type_ids", help="the type ids")
        gather_index = StrField(
            value="",
            help='if the gather_index is not "", we will gather the output based gather index',
        )

    input_map = NestField(
        value=InputMap, help="the input map of the static embedding module"
    )

    class OutputMap:
        embedding = StrField(value="embedding", help="the embedding")

    output_map = NestField(
        value=OutputMap, help="the output map of the static embedding module"
    )


@register("embedding", "bert_like")
class BertLikeEmbedding(SimpleModule):
    """Wrap the hugingface transformers"""

    def __init__(self, config: BertLikeEmbeddingConfig):
        super(BertLikeEmbedding, self).__init__(config)
        self.config = config
        self.pretrained_transformers = BertLike(config)

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        self.pretrained_transformers.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """get the transformers output as embedding

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        input_ids = (
            inputs[self.config.input_map.input_ids]
            if self.config.input_map.input_ids
            else None
        )
        attention_mask = (
            inputs[self.config.input_map.attention_mask]
            if self.config.input_map.attention_mask
            else None
        )
        type_ids = (
            inputs[self.config.input_map.type_ids]
            if self.config.input_map.type_ids
            else None
        )
        (
            sequence_output,
            all_hidden_states,
            all_self_attentions,
        ) = self.pretrained_transformers(
            {
                "input_ids": input_ids,
                "attention_mask": attention_mask,
                "token_type_ids": type_ids,
            }
        )
        if self.config.input_map.gather_index:
            gather_index = inputs[self.config.input_map.gather_index]
            g_bs, g_seq_len = gather_index.shape
            bs, seq_len, hid_size = sequence_output.shape
            assert g_bs == bs
            assert g_seq_len <= seq_len
            sequence_output = torch.gather(
                sequence_output[:, :, :],
                1,
                gather_index.unsqueeze(-1).expand(bs, g_seq_len, hid_size),
            )
        inputs[self.config.output_map.embedding] = sequence_output
        return inputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import pickle as pkl
from typing import Callable, Dict, Iterator, Type

import pandas as pd
import pyarrow.parquet as pq
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

logger = logging.getLogger(__name__)


@cregister("processor", "default")
class DefaultProcessorConfig(Base):
    """the default processor"""

    data_root = StrField(
        value=None,
        additions=[None],
        help="the root path of input(should be processed) data, if set the data_root to not null, all the path of data will be relative to the data_root",
    )
    train_data_type = StrField(
        value="none",
        options=[
            "dict",
            "dataframe",
            "parquet",
            "parquet_list",
            "json",
            "none",
            "pickle",
        ],
        help="the type of train data, `none` means no train data, `parquet_list` means the train data is a list of parquet file path, we will separate load->process->save the data into several parts, `pickle` means the data is a pickled `dataframe` file, we will load it directly.",
    )
    valid_data_type = StrField(
        value="none",
        options=["dict", "dataframe", "parquet", "json", "none"],
        help="the type of valid data, `none` means no valid data",
    )
    test_data_type = StrField(
        value="none",
        options=["dict", "dataframe", "parquet", "json", "none"],
        help="the type of test data, `none` means no test data",
    )
    predict_data_type = StrField(
        value="none",
        options=["dict", "dataframe", "parquet", "parquet_list", "json", "none"],
        help="the type of predict data, `none` means no predict data",
    )
    online_data_type = StrField(
        value="none",
        options=["dict", "dataframe", "none"],
        help="the type of online data, `none` means no online data",
    )
    feed_order = ListField(
        value=[],
        suggestions=[["tokenizer", "token_gather", "label_to_id", "token_embedding"]],
        help="the order of data feed",
    )
    meta_collection_on_train = BoolField(
        value=True,
        help="whether to collect meta info on train data, if there are more than one train data, we will only collect meta info on the first part.",
    )
    load_meta_on_start = BoolField(
        value=False,
        help="whether to load meta info when start the processor, when `load_meta_on_start` set to `True`, the `meta_collection_on_train` must be `False`.",
    )
    processed_data_dir = StrField(
        value="data/processed_data",
        help="the save dir of the processor, not effective by `data_root`.",
    )
    meta_dir = StrField(
        value="data/meta_data",
        help="the save dir of the meta info, not effective by `data_root`.",
    )
    do_save = BoolField(
        value=True,
        help="""
        whether save the processed data, 
        if `false` will return the processed dict
        """,
    )

    submodule = SubModule(
        value={},
        help="subprocessors for processor",
    )


def yield_dataframe(origin, data_type, config: DefaultProcessorConfig):
    if data_type == "none":
        yield None
    elif data_type == "dataframe":
        yield origin
    elif data_type == "dict":
        yield pd.DataFrame(data=origin)
    elif data_type == "pickle":
        assert isinstance(origin, str)
        if config.data_root:
            origin = os.path.join(config.data_root, origin)
        yield pkl.load(open(origin, "rb"))
    elif data_type == "json":
        assert isinstance(origin, str)
        if config.data_root:
            origin = os.path.join(config.data_root, origin)
        yield pd.read_json(origin)
    elif data_type == "parquet":
        assert isinstance(origin, str)
        if config.data_root:
            origin = os.path.join(config.data_root, origin)
        yield pq.read_table(origin).to_pandas()
    elif data_type == "parquet_list":
        assert isinstance(origin, list)
        for path in origin:
            if config.data_root:
                path = os.path.join(config.data_root, path)
            yield pq.read_table(path).to_pandas()
    else:
        raise NotImplementedError


@register("processor", "default")
class DefaultProcessor(object):
    """docstring for IProcessor"""

    stage_data_set_map = {
        "train": "train_data_set",
        "predict": "predict_data_set",
        "online": "online_data_set",
    }

    def __init__(self, stage: str, config: DefaultProcessorConfig):
        super(DefaultProcessor, self).__init__()
        config_dict = config._to_dict()
        self.stage = stage
        self.config: DefaultProcessorConfig = config

        assert (
            (not self.config.meta_collection_on_train)
            or (not self.config.load_meta_on_start)
            or (not self.config.train_data_type == "none")
        )

        self.subprocessors = {}
        droped_subprocessors = []
        self.will_processed_data_set = set()
        for name in self.config.feed_order:
            subprocessor_config_dict = config_dict[f"@subprocessor@{name}"]
            if not subprocessor_config_dict[self.stage_data_set_map[stage]]:
                logger.info(f"Skip '{name}' ....")
                droped_subprocessors.append(name)
                continue
            self.will_processed_data_set.update(
                set(subprocessor_config_dict[self.stage_data_set_map[stage]])
            )
            logger.info(f"Init '{name}' ....")
            subprocessor_name = subprocessor_config_dict["_name"].split("-")[0]
            subprocessor_config = cregister.get("subprocessor", subprocessor_name)(
                subprocessor_config_dict
            )
            subprocessor = register.get("subprocessor", subprocessor_name)(
                stage=self.stage,
                config=subprocessor_config,
                meta_dir=self.config.meta_dir,
            )
            if self.config.load_meta_on_start:
                subprocessor.load_meta()
            self.subprocessors[name] = subprocessor
        for name in droped_subprocessors:
            self.config.feed_order.remove(name)

    def load_data(self, data: Dict, type_name: str) -> Iterator:
        """load data
        Returns:
            Iterable DataFrame
        """
        if type_name == "train":
            return yield_dataframe(
                data.get("train", {}), self.config.train_data_type, self.config
            )
        if type_name == "valid":
            return yield_dataframe(
                data.get("valid", {}), self.config.valid_data_type, self.config
            )
        if type_name == "test":
            return yield_dataframe(
                data.get("test", {}), self.config.test_data_type, self.config
            )
        if type_name == "predict":
            return yield_dataframe(
                data.get("predict", {}), self.config.predict_data_type, self.config
            )
        raise NotImplementedError

    def save(self, data: pd.DataFrame, type_name: str, i: int):
        """save data to self.config.processed_data_dir

        Args:
            data: should saved data

        Returns:
            None
        """
        os.makedirs(
            os.path.join(self.config.processed_data_dir, type_name), exist_ok=True
        )
        data_path = os.path.join(self.config.processed_data_dir, type_name, f"{i}.pkl")
        assert (
            i == 0
        ), f"Currently only support save one {type_name} data"  # TODO: support save multi data
        pkl.dump(data, open(data_path, "wb"))

    def process(self, data: Dict) -> Dict:
        """Process entry

        Args:
            data:
            >>> {
            >>>     "train": {training data....},
            >>>     "test": ..
            >>> }

        Returns:
            processed data
        """
        result = {}
        if not self.config.do_save:
            result = {key: [] for key in data}
        for type_name in ["train", "valid", "test", "predict"]:
            if type_name not in self.will_processed_data_set:
                continue
            for i, loaded_data in enumerate(self.load_data(data, type_name)):
                if loaded_data is None:
                    continue
                deliver_meta = (
                    self.config.meta_collection_on_train
                    and type_name == "train"
                    and i == 0
                )
                for name in self.config.feed_order:
                    logger.info(
                        f"Processing on {type_name} {i if i > 0 else ''}: {name}"
                    )
                    loaded_data = self.subprocessors[name].process(
                        data=loaded_data, deliver_meta=deliver_meta
                    )
                if not self.config.do_save:
                    result[type_name].append(loaded_data)
                else:
                    self.save(loaded_data, type_name, i)
        return result

    def online_process(self, data: pd.DataFrame):
        """online server process the data without save
        Args:
            data:
                the data to be processed

        Returns:
            processed data
        """
        if "online" not in self.will_processed_data_set:
            return data
        for name in self.config.feed_order:
            data = self.subprocessors[name].process(data=data, deliver_meta=False)
        return data



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os

from dlk.utils.logger import logfile

if os.environ.get("DISABLE_LOGFILE", "0") not in {
    "1",
    "True",
    "true",
    "TRUE",
    "YES",
    "yes",
    "Yes",
} and os.environ.get("IN_INTC", "0") not in {"1", "True", "true", "TRUE", 1}:
    logfile()


from intc import cregister

from dlk.utils.register import register
